import numpy as np, os, warnings
from . import enmap, utils, wcsutils

def zeros(tile_geom, dtype=np.float64):
	"""Construct a zero-initialized TileMap with the given TileGeometry and data type"""
	flat = np.zeros(tile_geom.pre + (np.sum(tile_geom.npixs[tile_geom.active]),), dtype)
	return TileMap(flat, tile_geom.copy())

def empty(tile_geom, dtype=np.float64):
	"""Construct a zero-initialized TileMap with the given TileGeometry and data type"""
	flat = np.empty(tile_geom.pre + (np.sum(tile_geom.npixs[tile_geom.active]),), dtype)
	return TileMap(flat, tile_geom.copy())

def full(tile_geom, val, dtype=np.float64):
	"""Construct a zero-initialized TileMap with the given TileGeometry and data type"""
	flat = np.full(tile_geom.pre + (np.sum(tile_geom.npixs[tile_geom.active]),), val, dtype)
	return TileMap(flat, tile_geom.copy())

def from_tiles(tiles, tile_geom):
	"""Construct a TileMap from a set of a full list of tiles, both active
	and inactive. Inactive tiles are indicated with None entries. The active
	information in tile_geom is ignored, as is the non-pixel part of tile_geom.shape,
	which is instead inferred from the tiles."""
	active_tiles = []
	active       = []
	for gi, tile in enumerate(tiles):
		if tile is None: continue
		active_tiles.append(tile)
		active.append(gi)
	return from_active_tiles(active_tiles, tile_geom.copy(active=active))

def from_active_tiles(tiles, tile_geom):
	"""Construct a TileMap from a list of active tiles that should match the
	active list in the provided tile geometry. The non-pixel part of tile_geom
	is ignored, and is instead inferred from the tile shapes."""
	if len(tiles) != tile_geom.nactive:
		raise ValueError("Wrong number of tiles passed. Expected %d but got %d" % (tile_geom.nactive, len(tiles)))
	if len(tiles) == 0: return zeros(tile_geom)
	data = np.concatenate([tile.reshape(tile.shape[:-2]+(-1,)) for tile in tiles],-1)
	return TileMap(data, tile_geom.copy(pre=data.shape[:-1]))

class TileMap(np.ndarray):
	"""Implements a sparse tiled map, as described by a TileGeometry. This is effectively
	a large enmap that has been split into tiles, of which only a subset is stored. This
	is implemented as a subclass of ndarray instead of a list of tiles to allow us to
	transparently perform math operations on it. The maps are stored stored as a single
	array with all tiles concatenated along a flattened pixel dimension, in the same
	order as in tile_geom.active.

	Example: A TileMap covering a logical area with shape (3,100,100) with (10,10) tiles
	and active tiles [7,5] will have a shape of (3,200=10*10*2) when accessed directly.
	When accessed through the .tiles view, .tiles[5] will return a view of an (3,10,10) enmap,
	as will .tiles[7]. For all other indices, .tiles[x] will return None. The same
	two tiles can be accessed as .active_tiles[1] and .active_tiles[0] respecitvely.

	Slicing the TileMap using the [] operator works. For all but the last axis, this
	does what you would expect. E.g. for the above example, tile_map[0].tiles[5] would
	return a view of a (10,10) enmap (so the first axis is gone). If the last axis,
	which represents a flattened view of the pixels and tiles, is sliced, then the
	returned object will be a plain numpy array.
	"""
	def __new__(cls, arr, tile_geom):
		"""Constructs a TileMap object given a raw array arr[...,totpix] and
		a tuple of geometries."""
		obj = np.asarray(arr).view(cls)
		obj.geometry  = tile_geom
		return obj
	def __array_finalize__(self, obj):
		if obj is None: return
		self.geometry = getattr(obj, "geometry", None)
	def __repr__(self):
		return "TileMap(%s,%s)" % (np.asarray(self), str(self.geometry))
	def __str__(self): return repr(self)
	def __array_wrap__(self, arr, context=None, return_scalar=False):
		# In the future need to support `return_scalar`, but that is seemingly
		# undocumented and not actually supported in numpy 2.0? So for now we
		# just ignore it.
		return TileMap(arr, self.geometry)
	def __getitem__(self, sel):
		# Split sel into normal and wcs parts.
		sel1, sel2 = utils.split_slice(sel, [self.ndim-1,1])
		if len(sel2) > 1:
			raise IndexError("too many indices")
		elif len(sel2) == 1:
			# Degrade to plain array if we index the last, special pixel/tile axis
			return np.ndarray.__getitem__(self, sel)
		else:
			res  = np.ndarray.__getitem__(self, sel)
			ogeo = self.geometry.copy()
			ogeo.shape = res.shape[:-1]+self.geometry.shape[-2:]
			return TileMap(res, ogeo)
	def __getslice__(self, a, b=None, c=None): return self[slice(a,b,c)]
	def contig(self): return TileMap(np.ascontiguousarray(self), self.geometry)
	@property
	def pre(self): return self.geometry.pre
	@property
	def ntile(self): return self.geometry.ntile
	@property
	def nactive(self): return self.geometry.nactive
	def copy(self, order='K'):
		return TileMap(np.copy(self,order), self.geometry.copy())
	@property
	def tiles(self):
		return TileView(self, active=False)
	@property
	def active_tiles(self):
		return TileView(self, active=True)
	def with_tiles(self, other, strict=False):
		"""If a and b are TileMaps with the same overall tiling but different
		active tile sets, then c = a.with_tiles(b) will make c a TileMap
		with the union of the active tiles of a and b and the data from a
		(new tiles are zero-initialized).

		If strict==True, then c will have exactly the active tiles of b,
		in exactly that order. Binary operations on strictly compatible arrays
		should be considerably faster."""
		try: active = other.geometry.active
		except AttributeError: active = _parse_active(other, self.ntile)
		if np.all(active == self.geometry.active):
			return self.copy()
		# Construct the new geometry
		if strict: new_geom = self.geometry.copy(active    =active)
		else:      new_geom = self.geometry.copy(add_active=active)
		# Construct the new array
		res = zeros(new_geom, dtype=self.dtype)
		# Copy over data. Not optimized
		for gi in res.geometry.active:
			ai = self.geometry.lookup[gi]
			if ai >= 0: res.tiles[gi] = self.active_tiles[ai]
		return res
	# Forward some properties of TileGeometry
	@property
	def active(self): return self.geometry.active
	@property
	def lookup(self): return self.geometry.lookup
	@property
	def nactive(self): return self.geometry.nactive
	@property
	def ntile(self): return self.geometry.ntile
	@property
	def tile_shape(self): return self.geometry.tile_shape
	# General methods
	def insert(self, imap, op=lambda a,b:b): return insert(self, imap, op=op)

class TileView:
	"""Helper class used to implement access to the individual tiles that make up a TileMap object"""
	def __init__(self, tile_map, active=True):
		self.tile_map = tile_map
		self.active   = active
		self.offs     = utils.cumsum(tile_map.geometry.npixs[tile_map.geometry.active], endpoint=True)
	@property
	def ndim(self): return self.tile_map.ndim+1 # ndim of logical full map
	@property
	def shape(self): return self.tile_map.geometry.shape # shape of logical full map
	def __len__(self):
		if self.active: return len(self.tile_map.geometry.active)
		else:           return self.tile_map.geometry.ntile
	def __getitem__(self, sel):
		"""Get a single tile or subset of a tile from the TileMap. The first
		entry in the slice must be an integer - general slicing in the tile axis is not
		supported, though it could be added. The rest of the indices can be anything an
		enmap will accept."""
		if isinstance(sel, int):
			# Optimize common use case by avoiding slice decoding.
			# Doesn't save much time, really. 0.1 ms when combined with the
			# sel2 check below
			i      = sel
			sel2   = ()
		else:
			sel1, sel2 = utils.split_slice(sel, [1,self.tile_map.ndim+2-1])
			if len(sel1) == 0: return self.tile_map
			i      = sel1[0]
		geo  = self.tile_map.geometry
		if self.active:
			ai, gi = i, geo.active[i]
		else:
			ai, gi = geo.lookup[i], i
			# Return None for inactive tiles since that's what so3g does.
			# But consider raising an exception instead.
			if ai < 0: return None
		if ai < 0 or ai >= self.tile_map.nactive:
			raise IndexError("Active tile index %d (global %d) is out of bounds for TileMap with %d active tiles" % (ai, gi, self.tile_map.nactive))
		tile_info = geo.tiles[gi]
		tile = enmap.ndmap(self.tile_map[...,self.offs[ai]:self.offs[ai+1]].reshape(self.tile_map.pre + tile_info.shape[-2:]), tile_info.wcs)
		# Apply any slicing of the tile itself
		if len(sel2) > 0: tile = tile[sel2]
		return tile
	def __setitem__(self, sel, val):
		"""Set a single tile or subset of a tile to the given value, which can be a number
		or a compatibly shaped array. The first entry in the slice must be an integer -
		general slicing in the tile axis is not supported, though it could be added.
		The rest of the indices can be anything an enmap will accept. This relies on
		slicing not producing copies, and may fail if the TileMap is not contiguous."""
		sel1, sel2 = utils.split_slice(sel, [1,self.tile_map.ndim+2-1])
		if len(sel1) == 0: return self.tile_map
		i    = sel1[0]
		geo  = self.tile_map.geometry
		if self.active:
			ai, gi = i, geo.active[i]
		else:
			ai, gi = geo.lookup[i], i
			# Return None for inactive tiles since that's what so3g does.
			# But consider raising an exception instead.
			if ai < 0: raise IndexError("Tile %d is not active" % gi)
		# This assumes that the slicing and reshaping won't cause copies. This should be the case if
		# self.tile_map is contiguous. To be robust I should detect if a copy would be made, and fall back
		# on something slower in that case
		tile_info = geo.tiles[gi]
		enmap.ndmap(self.tile_map[...,self.offs[ai]:self.offs[ai+1]].reshape(self.tile_map.pre + tile_info.shape[-2:]), tile_info.wcs)[sel2] = val
	def __iter__(self):
		"""Iterator access. Faster than __getitem__ due to not having to resolve
		the more complicated slicing it supports."""
		geo = self.tile_map.geometry
		if self.active: items = [(ai,geo.active[ai]) for ai in range(geo.nactive)]
		else:           items = [(geo.lookup[gi],gi) for gi in range(geo.ntile)]
		for ai, gi in items:
			if ai < 0:
				yield None
			else:
				tile_info = geo.tiles[geo.active[ai]]
				yield enmap.ndmap(self.tile_map[...,self.offs[ai]:self.offs[ai+1]].reshape(self.tile_map.pre + tile_info.shape[-2:]), tile_info.wcs)

# Math operations. These are automatically supported by virtue of being a numpy subclass,
# but the functions below add support for expanding the active tiles automatically when TileMaps
# with incompatible active tiles are combined.
def make_binop(op, is_inplace=False):
	if isinstance(op, str):
		op = getattr(np.ndarray, op)
	def binop(self, other):
		if isinstance(other, TileMap): # could be replaced with a try statement for duck typing
			comp = self.geometry.compatible(other.geometry)
			if comp == 0:
				# Not compatible
				raise ValueError("operands could not be broadcast together with geometries %s and %s" % (str(self.geometry), str(other.geometry)))
			elif comp == 1:
				# Loosely compatible. Requires manual work. Slow, both due to looping and
				# poor memory localization.
				if is_inplace:
					# We can't change the shape of the output array for an in-place operation
					if np.any(self.geometry.lookup[other.geometry.active]<0):
						raise ValueError("Cannot expand TileMap with geometry %s to include active tiles in geometry %s in in-place operation" % (str(self.geometry), str(other.geometry)))
					opre = utils.broadcast_shape(self.pre, other.pre)
					if opre != self.pre:
						raise ValueError("operands could not be broadcast together with geometries %s and %s" % (str(self.geometry), str(other.geometry)))
					for gi in other.geometry.active:
						self.tiles[gi] = op(self.tiles[gi], other.tiles[gi])
					return self
				else:
					# Not in-place, so we have more flexibility. First build the output map
					oact = np.unique(np.concatenate([self.geometry.active, other.geometry.active]))
					otype= np.result_type(self.dtype, other.dtype)
					oshape = utils.broadcast_shape(self.pre, other.pre) + self.geometry.shape[-2:]
					ogeo = geometry(oshape, self.geometry.wcs, tile_shape=self.geometry.tile_shape, active=oact)
					out  = zeros(ogeo, otype)
					# Copy over our old values
					for gi in self.geometry.active:
						out.tiles[gi] = self.tiles[gi]
					# Then update with valus from other
					for gi in other.geometry.active:
						out.tiles[gi] = op(out.tiles[gi], other.tiles[gi])
					return out
			else:
				# Fully compatible. Handle outside
				pass
		# Handle fully compatible or plain array. Fast.
		out =  op(self, other)
		out = TileMap(out, self.geometry.copy(pre=out.shape[:-1]))
		return out
	return binop

for op in ["__add__", "__sub__", "__mul__", "__pow__", "__truediv__", "__floordiv__",
		"__lt__", "__le__", "__eq__", "__ne__", "__ge__", "__gt__",
		"__and__", "__or__", "__xor__", "__lshift__", "__rshift__"]:
	setattr(TileMap, op, make_binop(op))
for op in ["__iadd__", "__isub__", "__imul__", "__ipow__", "__itruediv__", "__ifloordiv__",
		"__iand__", "__ior__", "__ixor__", "__ilshift__", "__irshift__"]:
	setattr(TileMap, op, make_binop(op, is_inplace=True))

def insert(omap, imap, op=lambda a,b:b):
	"""Insert imap into omap, returning the result. Equivalent to enmap.insert, but with the
	following important differences:
	* omap is not modified. Use the result is returned. (enmap both modifies and returns)
	* The maps must have the same geometry, only differing by the active tiles.
	This may be generalized in the future."""
	binop = make_binop(op)
	return binop(omap, imap)

def map_mul(mat, vec):
	"""Elementwise matrix multiplication mat*vec. Result will have
	the same shape as vec. Multiplication happens along the last non-pixel
	indices."""
	# Allow scalar product, broadcasting if necessary
	mat = np.asanyarray(mat)
	if mat.ndim <= 2: return mat*vec
	# Otherwise we do a matrix product along the last axes
	ovec = samegeo(np.einsum("...abi,...bi->...ai", mat, vec), mat, vec)
	return ovec

def samegeo(arr, *args):
	"""Returns arr with the same geometry information as the first tilemap among
	args. If no matches are found, arr is returned as is.  Will
	reference, rather than copy, the underlying array data
	whenever possible.
	"""
	for m in args:
		try: return TileMap(arr, m.geometry.copy(pre=arr.shape[:-1]))
		except AttributeError: pass
	return arr

########################################
############ Geometry stuff ############
########################################

def geometry(shape, wcs, tile_shape=(500,500), active=[]):
	"""TileGeometry constructor.
	shape, wcs: The enmap geometry of the full space the tiling covers.
	tile_shape: The (ny,nx) vertical and horizontal tile shape in pixels.
	active:     The list of active tile indices."""
	shape      = tuple(shape)
	wcs        = wcs
	tile_shape = tuple(np.zeros(2,int)+tile_shape) # broadcast to len(2)
	# Get the grid shape by rounding up to include any partial tiles at the edge
	grid_shape = tuple([(s+ts-1)//ts for s,ts in zip(shape[-2:], tile_shape[-2:])])
	ntile      = grid_shape[0]*grid_shape[1]
	# Compute all the individual tile shapes. Should maybe be done on-the-fly
	tile_shapes= np.zeros(grid_shape + (2,), int)
	tile_shapes[:, :] = tile_shape
	tile_shapes[-1,:,0] = np.minimum(tile_shape[-2], (shape[-2]-1)%tile_shape[-2]+1)
	tile_shapes[:,-1,1] = np.minimum(tile_shape[-1], (shape[-1]-1)%tile_shape[-1]+1)
	tile_shapes = tile_shapes.reshape(-1,2)
	# Number of pixels per tile
	npixs       = tile_shapes[:,0]*tile_shapes[:,1]
	# And the active tile list and reverse lookup
	active      = np.array(active,int)
	lookup      = np.full(ntile,-1,int)
	lookup[active] = np.arange(len(active))
	# Call the raw constructor
	return TileGeometry(shape, wcs, tile_shape, grid_shape, tile_shapes, npixs, active, lookup)

class TileGeometry:
	def __init__(self, shape, wcs, tile_shape, grid_shape, tile_shapes, npixs, active, lookup):
		"""Raw constructor for a TileGeometry. You normally don't want to use this. Use
		tilemap.geometry() instead."""
		self.shape      = shape
		self.wcs        = wcs
		self.tile_shape = tile_shape
		self.grid_shape = grid_shape
		self.ntile      = self.grid_shape[0]*self.grid_shape[1]
		self.tile_shapes= tile_shapes
		self.npixs      = npixs
		self.active     = active
		self.lookup     = lookup
	def grid2ind(self, ty, tx):
		"""Get the index of the tile wiht grid coordinates ty,tx in the full tiling"""
		return ty*self.grid_shape[1]+tx
	def ind2grid(self, i):
		"""Get the tile grid coordinates ty, tx for tile #i in the full tiling"""
		nx = self.grid_shape[-1]
		return i//nx, i%nx
	def copy(self, pre=None, active=None, add_active=None):
		if pre is not None: shape = tuple(pre) + self.shape[-2:]
		else:               shape = self.shape
		_active = self.active.copy()
		lookup  = self.lookup.copy()
		if active is not None or add_active is not None:
			# Allow us to override these, which will require recalculation of lookup
			if active is not None:
				_active = _parse_active(active, self.ntile)
				lookup  = np.full(self.ntile,-1,int)
				lookup[_active] = np.arange(len(_active))
			if add_active is not None:
				add_active = _parse_active(add_active, self.ntile)
				_active = np.concatenate([_active, add_active[lookup[add_active]<0]])
				lookup[_active] = np.arange(len(_active))
		return TileGeometry(shape, self.wcs, self.tile_shape, self.grid_shape, self.tile_shapes.copy(), self.npixs.copy(), _active, lookup)
	@property
	def pre(self): return self.shape[:-2]
	@property
	def nactive(self): return len(self.active)
	@property
	def size(self): return np.prod(self.pre)*np.sum(self.npixs[self.active])
	@property
	def tiles(self):
		"""Allow us to get the enmap geometry of tile #i by writing
		tile_geom.tiles[i]"""
		return _TileGeomHelper(self)
	def __repr__(self): return "TileGeometry(%s, %s, tile_shape=%s, active=%s)" % (str(self.shape), str(self.wcs), str(self.tile_shape), str(self.active))
	def compatible(self, other):
		"""Return our compatibility with binary operations with other.
		The return value can be 2, 1 or 0:
		2: Strictly compatible. Both the logical geometry (shape, wcs), tile shape
		   and active tiles match. This allows for direct numpy operations without
		   any manual looping over tiles.
		1: Loosely compatible. The logical and tile geometry match, but not the active tiles.
		0. Not compatible."""
		if tuple(self.shape[-2:]) != tuple(other.shape[-2:]): return 0
		if tuple(self.tile_shape) != tuple(other.tile_shape): return 0
		if self.nactive == other.nactive and np.all(self.active == other.active): return 2
		else: return 1

class _TileGeomHelper:
	def __init__(self, tile_geom):
		self.tile_geom = tile_geom
	def __getitem__(self, i):
		"""Return the geometry of index #i in the full tiling"""
		g  = self.tile_geom
		ty, tx = g.ind2grid(i)
		y1 = ty*g.tile_shape[-2]
		x1 = tx*g.tile_shape[-1]
		y2 = min(y1+g.tile_shape[-2], g.shape[-2])
		x2 = min(x1+g.tile_shape[-1], g.shape[-1])
		return enmap.Geometry(g.shape, g.wcs)[...,y1:y2,x1:x2]

def _parse_active(active, ntile):
	if utils.streq(active, "all"): return np.arange(ntile,dtype=int)
	else: return np.asarray(active,int)

def to_enmap(tile_map):
	omap = enmap.zeros(tile_map.geometry.shape, tile_map.geometry.wcs, tile_map.dtype)
	for ai, tile in enumerate(tile_map.active_tiles):
		gi     = tile_map.active[ai]
		gy, gx = tile_map.geometry.ind2grid(gi)
		th, tw = tile_map.geometry.tile_shape
		y1     = gy*th
		y2     = min((gy+1)*th, omap.shape[-2])
		x1     = gx*tw
		x2     = min((gx+1)*tw, omap.shape[-1])
		omap[...,y1:y2,x1:x2] = tile
	return omap

############################################
########## Distributed TileMaps ############
############################################

# The main purpose of a TileMap is to spead the data of a huge map across many mpi tasks.

def redistribute(imap, comm, active=None, omap=None):
	"""Redistirbute the data in the mpi-distributed tiles in imap into the
	active tiles in omap, using the given communicator. If a tile is active in
	multiple tasks in imap, it will be reduced. If it is active in multiple tiles in
	omap, it will be duplicated."""
	# 1. Who owns what?
	iactive     = np.zeros(imap.ntile,bool); iactive[imap.active]=True
	iactive_all = utils.allgather(iactive, comm) # [ntask,ntile]
	# 2. Who should own what? Determine automatically if not given
	if omap is None:
		if active is None:
			iactive_any = np.nonzero(np.any(iactive_all,0))[0]
			active = np.array_split(iactive_any, comm.size)[comm.rank]
		omap = zeros(imap.geometry.copy(active=active), dtype=imap.dtype)
	oactive     = np.zeros(omap.ntile,bool); oactive[omap.active]=True
	oactive_all = utils.allgather(oactive, comm) # [ntask,ntile]
	# 3. Figure out who I should send and receive each of my tiles to/from
	omask     = oactive_all[:,imap.active] # [ntasks,iactive]
	imask     = iactive_all[:,omap.active] # [ntasks,oactive]
	isizes    = imap.geometry.npixs[imap.active] * np.prod(imap.pre).astype(int)
	osizes    = omap.geometry.npixs[omap.active] * np.prod(omap.pre).astype(int)
	ioffs     = utils.cumsum(isizes)
	ooffs     = utils.cumsum(osizes)
	# 4. Build our alltoallv send info
	iflat     = np.ascontiguousarray(imap.T).reshape(-1)
	send_sizes= np.sum(omask*isizes,1)
	send_offs = utils.cumsum(send_sizes)
	send_buf  = [iflat[ioffs[iact]:ioffs[iact]+isizes[iact]] for rank, iact in np.argwhere(omask)]
	send_buf  = np.concatenate(send_buf, dtype=iflat.dtype) if len(send_buf) > 0 else np.zeros(0, iflat.dtype)
	# 5. Build the alltoallv receive info
	recv_sizes= np.sum(imask*osizes,1)
	recv_offs = utils.cumsum(recv_sizes)
	recv_buf  = np.zeros(np.sum(recv_sizes), omap.dtype)
	# 6. Perform the actual communication
	comm.Alltoallv((send_buf, (send_sizes, send_offs)), (recv_buf, (recv_sizes, recv_offs)))
	del iflat, send_buf
	# 7. Copy and reduce into flattened output tiles
	oflat     = np.zeros(omap.size, omap.dtype)
	rbuf_off  = 0
	for rank, oact in np.argwhere(imask):
		oflat[ooffs[oact]:ooffs[oact]+osizes[oact]] += recv_buf[rbuf_off:rbuf_off+osizes[oact]]
		rbuf_off += osizes[oact]
	del recv_buf
	# 8. And move data into our actual output map
	omap[:] = oflat.reshape(omap.shape[::-1]).T
	return omap

def tree_reduce(imap, comm, plan=None):
	"""Given a tilemap imap that's distributed over the communicator comm,
	and where each tile is potentially present in multiple tasks, sum the
	duplicate tiles and assign them to a single task, such that in the end
	each tile is present in at most one task. Exactly which tiles end up in
	which tasks is determined automatically but deterministically based on the
	tile ownership pattern in imap."""
	from map_reduce import distlib
	if plan is None:
		plan = distlib.Logistics(imap.active, comm)
	work = [None if tile is None else tile.copy() for tile in imap.tiles]
	for gi, sender, receiver in plan.ops:
		if comm.rank == sender:
			comm.Send(work[gi], dest=receiver)
			work[gi] = None
		elif comm.rank == receiver:
			tile = np.zeros_like(work[gi])
			comm.Recv(tile, source=sender)
			work[gi] += tile
	omap = from_tiles(work, imap.geometry)
	return omap

def get_active_distributed(tile_map, comm):
	# Figure out which tiles are owned by anybody
	iactive     = np.zeros(tile_map.ntile,int); iactive[tile_map.active]=1
	iactive     = utils.allreduce(iactive, comm)
	return np.nonzero(iactive)[0]

def reduce(tile_map, comm, root=0):
	"""Given a distributed TileMap tile_map, collect all the tiles
	on the task with rank root (default is rank 0), and return it.
	Multiply owned tiles are reduced. Returns a TileMap with no
	active tiles for other tasks than root."""
	active_distributed = get_active_distributed(tile_map, comm)
	active = active_distributed if comm.rank == root else []
	return redistribute(tile_map, comm, active)

def write_map(fname, tile_map, comm, extra={}):
	"""Write a distributed tile_map to disk as a single enmap.
	Collects all the data on a single task before writing."""
	omap = reduce(tile_map, comm)
	if comm.rank == 0:
		omap = to_enmap(omap)
		enmap.write_map(fname, omap, allow_modify=True, extra=extra)



from __future__ import division, print_function
import numpy as np, argparse, time, sys, warnings, os, shlex, glob, PIL.Image, PIL.ImageDraw, fractions
from scipy import ndimage
from . import enmap, colorize, mpi, cgrid, utils, memory, bunch, wcsutils
# Optional dependency array_ops needed for contour drawing
try: from . import array_ops
except ImportError: pass

# Python 3 compatibility
try: basestring
except NameError: basestring = str

class Printer:
	def __init__(self, level=1, prefix=""):
		self.level  = level
		self.prefix = prefix
	def write(self, desc, level, exact=False, newline=True, prepend=""):
		if level == self.level or not exact and level <= self.level:
			prepend = "%6.2f " % (memory.max()/1024.**3) + prepend
			sys.stderr.write(prepend + self.prefix + desc + ("\n" if newline else ""))
	def push(self, desc):
		return Printer(self.level, self.prefix + desc)
	def time(self, desc, level, exact=False, newline=True):
		class PrintTimer:
			def __init__(self, printer): self.printer = printer
			def __enter__(self): self.time = time.time()
			def __exit__(self, type, value, traceback):
				self.printer.write(desc, level, exact=exact, newline=newline, prepend="%6.2f " % (time.time()-self.time))
		return PrintTimer(self)
noprint = Printer(0)

def plot(*arglist, **args):
	"""The main plotting function in this module. Plots the given maps/files,
	returning them as a list of plots, one for each separate image. This
	function has two equivalent interfaces:
	1. A command-line like interface, where plotting options are specified with
	strings like "-r 500:50 -m 0 -t 2".
	2. A python-like interface, where plotting options are specified with
	keyword arguments, like range="500:50", mask=0, ticks=2.
	These interfaces can be mixed and matched.
	
	Input maps are specified either as part of the option strings, as separate
	file names, or as enmaps passed to the function. Here are some examples:
	
	plot(mapobj):
	  Plots the given enmap object mapobj. If mapobj is a scalar map,
	  the a length-1 list containing a single plot will be returned.
	  If mapobj is e.g. a 3-component TQU map, then a length-3 list
	  of plots will be returned.
	plot((mapobj,"foo")):
	  If a tuple is given, the second element specifies the name tag to
	  use. This tag will be used to populate the plot.name attribute
	  for each output plot, which can be useful when plotting and writing
	  the maps.
	 plot("file.fits"):
	  Reads a map from file.fits, and plots it. This sets the tag to "file",
	  so that the result can easily be written to "file.png" (or "file_0.png" etc).
	 plot(["a.fits","b.fits",(mapobj,"c"),(mapobj,"d.fits")])
	  Reads a.fits and plots it to a.png, reads b.fits and plots it to b.png,
	  plots the first mapobj to c.png and the second one to d.png (yes, the
	  extension in the filename specified in the tuple is ignored. This is
	  because that filename actually supplies the *input* filename that the
	  output filename should be computed from).
	 plot("foo*.fits")
	  Reads and plots every file matching the glob foo*.fits.
	 plot("a.fits -r 500:50 -m 0 -d 4 -t 4")
	  Reads and plots the file a.fits to a.png, using a color range of +-500
	  for the first field in the map (typically the total intensity), and
	  +-50 for the remaining fields (typically Q and U). The maps are downsampled
	  by a factor of 4, and plotted with a grid spacing of 4.
	
	Here is a list of the individual options plot recognizes. The short and long
	ones are recognized when used in the argument string syntax, while the long
	ones (with - replaced by _) also work as keyword arguments.

	See plot_iterator for an iterator version of this function.
	"""
	return list(plot_iterator(*arglist, **args))

def pshow(*arglist, method="auto", **args):
	"""Convenience function to both build plots and show them.
	pshow(...) is equivalent to show(plot(...))."""
	show(plot(*arglist, **args), method=method)

def pwrite(*arglist, **args):
	"""Convenience function to both build plots and write them.
	pwrite(...) is equivalent to write(plot(...))."""
	write(plot(*arglist, **args))

# Compatibility function
def get_plots(*arglist, **args):
	"""This function is identical to enplot.plot"""
	return plot(*arglist, **args)

def plot_iterator(*arglist, **kwargs):
	"""Iterator that yields the plots for each input map/file. Each yield
	will be a plot object, with fields
	.type: The type of the plot. Can be "pil" or "mpl". Usually "pil".
	.img:  The plot image object, of the given .type.
	.name: Suggested file name
	These plots objects can be written to disk using enplot.write.
	See the plot function documentation for a description of the arguments"""
	# This allows us to pass in both command-line style plot specifications
	# as well as python-style lists of enmaps and keyword arguments. The maps
	# to be plotted are collected in maps, which is a list that can contain
	# either enmaps or filenames. Filenames will later be read in as enmaps
	imaps = []
	comm   = extract_arg(kwargs, "comm",   None)
	noglob = extract_arg(kwargs, "noglob", False)

	# Set up defaults
	args = parse_args([])
	# Then process all the args
	for arg in arglist:
		if isinstance(arg, basestring):
			parsed = parse_args(arg, noglob=noglob, nodef=True)
			imaps += parsed.ifiles
			args.update(parsed)
		else:
			imaps.append(arg)
	del args["ifiles"]
	# Override with kwargs
	args.update(kwargs)
	args = bunch.Bunch(**args)
	# Check for invalid kwargs
	check_args(args)
	if comm is None:
		comm = mpi.COMM_SELF
	# Set up verbose output
	printer = Printer(args.verbosity)
	cache = {}
	# Plot each map
	for fi in range(comm.rank,len(imaps),comm.size):
		imap  = imaps[fi]
		if isinstance(imap, basestring):
			iname = imap
		elif isinstance(imap, tuple):
			imap, iname = imap
		else:
			iname = ""
			#if len(imaps) == 1: iname = "map.fits"
			#else: iname = "map%0*d.fits" % (get_num_digits(len(imaps)), fi)
		with printer.time("read %s" % iname, 3):
			map, minfo = get_map(imap, args, return_info=True, name=iname)
		with printer.time("ranges", 3):
			crange= get_color_range(map, args)
		for ci, cr in enumerate(crange.T):
			printer.write("color range %d: %12.5e to %15.7e" % (ci, cr[0], cr[1]), 4)
		# Loop over map fields
		ncomp  = map.shape[0]
		gsize  = 3 if args.rgb else 1
		ngroup = ncomp//gsize
		# Collect output images for this map in a list
		for gi in range(ngroup):
			# The unflattened index of the current field
			N = minfo.ishape[:-2]
			I = np.unravel_index(gi, N) if len(N) > 0 else []
			if args.symmetric and np.any(np.sort(I) != I):
				continue
			# Construct default out format
			ndigit   = get_num_digits(ngroup)
			ndigits  = [get_num_digits(n) for n in N]
			subprint = printer.push(("%%0%dd/%%d " % ndigit) % (gi+1,ncomp))
			dir, base, ext = split_file_name(minfo.fname)
			if args.odir is not None: dir = args.odir
			map_field = map[gi*gsize:(gi+1)*gsize]
			if minfo.wcslist:
				# HACK: If stamp extraction messed with the wcs, fix it here
				map_field.wcs = minfo.wcslist[I[0]]
			# We want field numbers in the output unless there's only one field, or
			# unless we're outputting a video
			is_vid    = is_video_ext(args.ext)
			want_inds = ngroup > 1 and not is_vid
			# Build output file name
			oinfo = {"dir":"" if dir == "." else dir + "/", "base":base, "iext":ext,
					"fi":fi, "fn":len(imaps), "ci":gi, "cn":ngroup, "pi":comm.rank, "pn":comm.size,
					"pre":args.prefix, "suf":args.suffix,
					"comp": "_"+"_".join(["%0*d" % (ndig,ind) for ndig,ind in zip(ndigits,I)]) if want_inds else "",
					"fcomp": "_%0*d" % (ndigit,gi) if len(minfo.ishape) > 2 else "",
					"ext":args.ext, "layer":""}
			oname = args.oname.format(**oinfo)
			# Draw the map
			if args.driver.lower() == "pil":
				img, info = draw_map_field(map_field, args, crange[:,gi*gsize:(gi+1)*gsize], return_info=True, return_layers=args.layers, printer=subprint, cache=cache)
				padding = np.array([-info.bounds[0,::-1],info.bounds[1,::-1]-map_field.shape[-2:]],dtype=int)
				printer.write("padded by %d %d %d %d" % tuple(padding.reshape(-1)), 4)
				typ = "vid" if is_vid else "pil"
				if args.layers:
					for layer, name in zip(img, info.names):
						oinfo["layer"] = "_" + name
						oname = args.oname.format(**oinfo)
						yield bunch.Bunch(img=layer, name=oname, type=typ, info=info, printer=subprint, **oinfo)
				else:
					yield bunch.Bunch(img=img, name=oname, type=typ, info=info, printer=subprint, **oinfo)
			elif args.driver.lower() in ["matplotlib","mpl"]:
				figure = draw_map_field_mpl(map_field, args, crange[:,gi*gsize:(gi+1)*gsize], printer=subprint)
				yield bunch.Bunch(img=figure, name=oname, type="mpl", dpi=args.mpl_dpi, printer=subprint, **oinfo)
			# Progress report
			printer.write("\r%s %5d/%d" % (iname, gi+1,ngroup), 2, exact=True, newline=False)
		printer.write("",    2, exact=True)
		printer.write(iname, 1, exact=True)

def write(fname, plots=None, writer=None):
	"""Write the given plot or plots to file. If plot is a single plot, then
	it will simply be written to the specified filename. If plot is a list of
	plots, then it fname will be interpreted as a prefix, and the plots will
	be written to prefix + plot.name for each individual plot. If name was
	specified during plotting, then plot.name will either be ".png" for scalar
	maps or "_0.png", "_1.png", etc. for vector maps. It's also possible to pass in
	plain images (either PIL or matplotlib), which will be written to the given
	filename."""
	if plots is None: fname, plots = "", fname
	# Make sure it's a list
	if not isinstance(plots, list) and not isinstance(plots, tuple):
		# Copy because we'll override the file name in this mode.
		# See the docstring
		plots  = [plots.copy()]
		plots[0].name = fname
		prefix = ""
	else:
		prefix = fname
	# Set up a Writer. We need this because for the the video output format,
	# multiple plots will be accumuated into a single output stream
	if writer is None:
		with PlotWriter() as writer:
			# Process each plot in the list
			for i, plot in enumerate(plots):
				_write_helper(writer, plot, prefix=prefix)
	else:
		for i, plot in enumerate(plots):
			_write_helper(writer, plot, prefix=prefix)

def _write_helper(writer, plot, prefix=""):
	# Normally we would have been given a plot object.
	# If someone passed in a plain img or fig, we need
	# to turn it into a plot
	if   isinstance(plot, bunch.Bunch): pass # already a proper plot
	elif isinstance(plot, PIL.Image):
		plot = bunch.Bunch(img=plot, type="pil", name="")
	else:
		# Assume matplotlib otherwise
		plot = bunch.Bunch(img=plot, type="mpl", name="", dpi=75)
	# Send it to the writer
	printer = plot.printer if "printer" in plot else noprint
	with printer.time("write to %s" % (prefix + plot.name), 3):
		writer.process(plot, prefix=prefix)

def define_arg_parser(nodefault=False):
	argdefs = []
	def add_argument(*args, default=None, **kwargs):
		short = [arg[1:] for arg in args if len(arg) >= 2 and arg[0] == "-" and arg[1] != "-"]
		long_ = [arg[2:] for arg in args if len(arg) >= 3 and arg[:2] == "--" and arg[2] != "-"]
		if len(long_) > 0: name = long_[0]
		else:              name = short[0]
		name = name.replace("-","_")
		if nodefault: default=argparse.SUPPRESS
		kwargs["default"] = default
		argdefs.append([name, [args, kwargs]])
	add_argument("-o", "--oname", default="{dir}{pre}{base}{suf}{comp}{layer}.{ext}", help="The format to use for the output name. Default is {dir}{pre}{base}{suf}{comp}{layer}.{ext}")
	add_argument("-c", "--color", default="planck", help="The color scheme to use, e.g. planck, wmap, gray, hotcold, etc., or a colors pecification in the form val:rrggbb,val:rrggbb,.... Se enlib.colorize for details.")
	add_argument("-r", "--range", type=str, help="The symmetric color bar range to use. If specified, colors in the map will be truncated to [-range,range]. To give each component in a multidimensional map different color ranges, use a colon-separated list, for example -r 250:100:50 would plot the first component with a range of 250, the second with a range of 100 and the third and any subsequent component with a range of 50.")
	add_argument("--min", type=str, help="The value at which the color bar starts. See --range.")
	add_argument("--max", type=str, help="The value at which the color bar ends. See --range.")
	add_argument("-q", "--quantile", type=float, default=0.01, help="Which quantile to use when automatically determining the color range. If specified, the color bar will go from [quant(q),quant(1-q)].")
	add_argument("-v", dest="verbosity", action="count", default=0, help="Verbose output. Specify multiple times to increase verbosity further.")
	add_argument("-u", "-s", "--upgrade", "--scale", type=str, default="1", help="Upscale the image using nearest neighbor interpolation by this amount before plotting. For example, 2 would make the map twice as large in each direction, while 4,1 would make it 4 times as tall and leave the width unchanged.")
	add_argument("--verbosity", dest="verbosity", type=int, help="Specify verbosity directly as an integer.")
	add_argument("--method", default="auto", help="Which colorization implementation to use: auto, fortran or python.")
	add_argument("--slice", type=str, help="Apply this numpy slice to the map before plotting.")
	add_argument("--sub",   type=str, help="Slice a map based on dec1:dec2,ra1:ra2.")
	add_argument("-H", "--hdu",  type=int, default=0, help="Header unit of the fits file to use")
	add_argument("--op", type=str, help="Apply this general operation to the map before plotting. For example, 'log(abs(m))' would give you a lograithmic plot.")
	add_argument("--op2", type=str, help="Like op, but allows multiple statements")
	add_argument("-d", "--downgrade", type=str, default="1", help="Downsacale the map by this factor before plotting. This is done by averaging nearby pixels. See --upgrade for syntax.")
	add_argument("--prefix", type=str, default="", help="Specify a prefix for the output file. See --oname.")
	add_argument("--suffix", type=str, default="", help="Specify a suffix for the output file. See --oname.")
	add_argument("--odir",   type=str, default=None, help="Override the output directory. See --oname.")
	add_argument("--ext", type=str, default="png", help="Specify an extension for the output file. This will determine the file type of the resulting image. Can be anything PIL recognizes. The default is png.")
	add_argument("-m", "--mask", type=float, help="Mask this value, making it transparent in the output image. For example -m 0 would mark all values exactly equal to zero as missing.")
	add_argument("--mask-tol", type=float, default=1e-14, help="The tolerance to use with --mask.")
	add_argument("-g", "--grid", action="count", default=1, help="Toggle the coordinate grid. Disabling it can make plotting much faster when plotting many small maps.")
	add_argument("--grid-color", type=str, default="00000020", help="The RGBA color to use for the grid.")
	add_argument("--grid-width", type=int, default=1, help="The line width to use for the grid.")
	add_argument("-t", "--ticks", type=str, default="1", help="The grid spacing in degrees. Either a single number to be used for both axis, or ty,tx.")
	add_argument("--tick-unit", "--tu", type=str, default=None, help="Units for tick axis. Can be the unit size in degrees, or the word 'degree', 'arcmin' or 'arcsec' or the shorter 'd','m','s'.")
	add_argument("--nolabels", action="store_true", help="Disable the generation of coordinate labels outside the map when using the grid.")
	add_argument("--nstep", type=int, default=200, help="The number of steps to use when drawing grid lines. Higher numbers result in smoother curves.")
	add_argument("--subticks", type=float, default=0, help="Subtick spacing. Only supported by matplotlib driver.")
	add_argument("-b", "--colorbar", default=0, action="count", help="Whether to draw the color bar or not")
	add_argument("--font", type=str, default="arial.ttf", help="The font to use for text.")
	add_argument("--font-size", type=int, default=20, help="Font size to use for text.")
	add_argument("--font-color", type=str, default="000000", help="Font color to use for text.")
	add_argument("-D", "--driver", type=str, default="pil", help="The driver to use for plotting. Can be pil (the default) or mpl. pil cleanly maps input pixels to output pixels, and has better coordiante system support, but doesn't have as pretty grid lines or axis labels.")
	add_argument("--mpl-dpi", type=float, default=75, help="The resolution to use for the mpl driver.")
	add_argument("--mpl-pad", type=float, default=1.6, help="The padding to use for the mpl driver.")
	add_argument("--rgb", action="store_true", help="Enable RGB mode. The input maps must have 3 components, which will be interpreted as red, green and blue channels of a single image instead of 3 separate images as would be the case without this option. The color scheme is overriden in this case.")
	add_argument("--rgb-mode", type=str, default="direct", help="The rgb mode to use. Can be direct or direct_colorcap. These only differ in whether colors are preserved when too high or low colors are capped. direct_colorcap preserves colors, at the cost of noise from one noisy component leaking into others during capping.")
	add_argument("--reverse-color",  action="store_true", help="Reverse the color scale. For example, a black-to-white scale will become a white-to-black sacle.")
	add_argument("-a", "--autocrop", action="store_true", help="Automatically crop the image by removing expanses of uniform color around the edges. This is done jointly for all components in a map, making them directly comparable, but is done independently for each input file.")
	add_argument("-A", "--autocrop-each", action="store_true", help="As --autocrop, but done individually for each component in each map.")
	add_argument("-L", "--layers", action="store_true", help="Output the individual layers that make up the final plot (such as the map itself, the coordinate grid, the axis labels, any contours and lables) as individual files instead of compositing them into a final image.")
	add_argument(      "--no-image", action="store_true", help="Skip the main image plotting. Useful for getting a pure contour plot, for example.")
	add_argument("-C", "--contours", type=str, default=None, help="Enable contour lines. For example -C 10 to place a contour at every 10 units in the map, -C 5:10 to place it at every 10 units, but starting at 5, and 1,2,4,8 or similar to place contours at manually chosen locations.")
	add_argument("--contour-type",  type=str, default="uniform", help="The type of the contour specification. Only used when the contours specification is a list of numbers rather than a string (so not from the command line interface). 'uniform': the list is [interval] or [base, interval]. 'list': the list is an explicit list of the values the contours should be at.")
	add_argument("--contour-color", type=str, default="000000", help="The color scheme to use for contour lines. Either a single rrggbb, a val:rrggbb,val:rrggbb,... specification or a color scheme name, such as planck, wmap or gray.")
	add_argument("--contour-width", type=int, default=1, help="The width of each contour line, in pixels.")
	add_argument("--annotate",      type=str, default=None, help="""Annotate the map with text, lines or circles. Should be a text file with one entry per line, where an entry can be:
		c[ircle] lat lon dy dx [rad [width [color]]]
		t[ext]   lat lon dy dx text [size [color]]
		l[ine]   lat lon dy dx lat lon dy dx [width [color]]
	dy and dx are pixel-unit offsets from the specified lat/lon.
	Alternatively, from python one can pass in a list of lists containig the same information, e.g. [["circle", 5.10,222.3,0,0,32,3,"black"]]""")
	add_argument("--annotate-maxrad", type=int, default=0, help="Assume that annotations do not extend further than this from their center, in pixels. This is used to prune which annotations to attempt to draw, as they can be a bit slow. The special value 0 disables this.")
	add_argument("--stamps", type=str, default=None, help="Plot stamps instead of the whole map. Format is srcfile:size:nmax, where the last two are optional. srcfile is a file with [ra dec] in degrees, size is the size in pixels of each stamp, and nmax is the max number of stamps to produce.")
	add_argument("--tile",  type=str, default=None, help="Stack components vertically and horizontally. --tile 5,4 stacks into 5 rows and 4 columns. --tile 5 or --tile 5,-1 stacks into 5 rows and however many columns are needed. --tile -1,5 stacks into 5 columns and as many rows are needed. --tile -1 allocates both rows and columns to make the result as square as possible. The result is treated as a single enmap, so the wcs will only be right for one of the tiles.")
	add_argument("--tile-transpose", action="store_true", help="Transpose the ordering of the fields when tacking. Normally row-major stacking is used. This sets column-major order instead.")
	add_argument("--tile-dims", type=str, default=None)
	add_argument("-S", "--symmetric", action="store_true", help="Treat the non-pixel axes as being asymmetric matrix, and only plot a non-redundant triangle of this matrix.")
	add_argument("-z", "--zenith",    action="store_true", help="Plot the zenith angle instead of the declination.")
	add_argument("-F", "--fix-wcs",   action="store_true", help="Fix the wcs for maps in cylindrical projections where the reference point was placed too far away from the map center.")
	add_argument(      "--pos-ra",    action="store_true", help="RA goes from 0 to 360 instead of -180 to 180")

	# Define the argument parser
	parser   = argparse.ArgumentParser()
	optnames = ["ifiles"]
	parser.add_argument("ifiles", nargs="*", help="The map files to plot. Each file will be processed independently and output as an image file with a name derived from that of the input file (see --oname). For each file a color range will be determined, and each component of the map (if multidimensional) will be written to a separate image file. If the file has more than 1 non-pixel dimension, these will be flattened first.")
	for name, (args, kwargs) in argdefs:
		parser.add_argument(*args, **kwargs)
		optnames.append(name)
	optnames = set(optnames)
	return parser, optnames

arg_parser, optnames = define_arg_parser()
arg_parser_nodef, _  = define_arg_parser(nodefault=True)
# Hack: Update the plot docstring. I suspect that this will confuse automated tools
help_short= "\n\t".join(arg_parser.format_help().split("positional arguments:")[0].split("\n")).rstrip()
try:
	help_long = "\n\t".join(arg_parser.format_help().split("optional arguments:")[1].split("\n"))
except IndexError:
	help_long = "\n\t".join(arg_parser.format_help().split("options:")[1].split("\n"))
plot.__doc__ += "\n\t" + help_long

def parse_args(args=sys.argv[1:], noglob=False, nodef=False):
	if isinstance(args, basestring):
		args = shlex.split(args)
	if nodef: res = arg_parser_nodef.parse_args(args)
	else:     res = arg_parser      .parse_args(args)
	res = bunch.Bunch(**res.__dict__)
	# Glob expansion
	if not noglob:
		ifiles = []
		for pattern in res.ifiles:
			matches = sorted(glob.glob(pattern))
			if len(matches) > 0:
				ifiles += matches
			else:
				ifiles.append(pattern)
		res.ifiles = ifiles
	return res

def extract_arg(args, name, default):
	if name not in args: return default
	res = args[name]
	del args[name]
	return res

def check_args(kwargs):
	for key in kwargs:
		if not key in optnames:
			msg = "Unrecognized argument '%s'\n\n%s" % (key, help_short)
			raise ValueError(msg)

def get_map(ifile, args, return_info=False, name=None):
	"""Read the specified map, and massage it according to the options
	in args. Relevant ones are sub, autocrop, slice, op, downgrade, scale,
	mask. Retuns with shape [:,ny,nx], where any extra dimensions have been
	flattened into a single one."""
	# TODO: this should be reorganized so that slicing can happen earlier.
	# Currently the whole file needs to be read.
	with warnings.catch_warnings():
		warnings.filterwarnings("ignore")
		if isinstance(ifile, basestring):
			toks  = ifile.split(":")
			ifile, slice = toks[0], ":".join(toks[1:])
			m0    = enmap.read_map(ifile, hdu=args.hdu, delayed=True)
			if name is None: name = ifile
		else:
			m0    = ifile
			slice = ""
			if name is None: name = ".fits"
		# This fills in a dummy, plain wcs if one does not exist
		try: m0.wcs
		except AttributeError: m0 = enmap.enmap(m0[:], copy=False)
		# Optionally fix the wcs to avoid crpix being too far away
		if args.fix_wcs:
			m0.wcs = wcsutils.fix_wcs(m0.wcs)
		# Save the original map, so we can compare its wcs later
		m  = m0
		# Submap slicing currently has wrapping issues
		if args.sub is not None:
			default = [[-90,-180],[90,180]]
			sub  = np.array([[(default[j][i] if q == '' else float(q))*np.pi/180 for j,q in enumerate(w.split(":"))]for i,w in enumerate(args.sub.split(","))]).T
			m = m.submap(sub)
		# Perform a common autocrop across all fields
		if args.autocrop:
			m = enmap.autocrop(m[:], value=0)
		# If necessary, split into stamps. If no stamp splitting occurs,
		# a list containing only the original map is returned
		mlist = extract_stamps(m, args)
		# The stamp stuff is a bit of an ugly hack. This loop and wcslist
		# are parts of that hack.
		for i, m in enumerate(mlist):
			# Downgrade
			downgrade = parse_list(args.downgrade, int)
			m = enmap.downgrade(m[:], downgrade)
			# Slicing, either at the file name level or though the slice option
			m = eval("m"+slice)
			if args.slice is not None:
				try: m = eval("m"+args.slice)
				except AttributeError: m = eval("m[:]"+args.slice) # handle proxy case
			# Unwrap any remaining proxy
			m    = m[:]
			flip = (m.wcs.wcs.cdelt*m0.wcs.wcs.cdelt)[::-1]<0
			assert m.ndim >= 2, "Image must have at least 2 dimensions"
			# Apply arbitrary map operations
			m1 = m
			if args.op is not None:
				m = eval(args.op, {"m":m,"enmap":enmap,"utils":utils,"np":np},np.__dict__)
			if args.op2 is not None:
				loc  = {"m":m}
				glob = {"enmap":enmap,"utils":utils,"np":np}
				exec(args.op2, glob, loc)
				m = loc["m"]
			# Scale if requested
			scale = parse_list(args.upgrade, int)
			if np.any(np.array(scale)>1):
				m = enmap.upgrade(m, scale)
			# Flip such that pixels are in PIL or matplotlib convention,
			# such that RA increases towards the left and dec upwards in
			# the final image. Unless a slicing operation on the image
			# overrrode this.
			if m.wcs.wcs.cdelt[1] > 0: m = m[...,::-1,:]
			if m.wcs.wcs.cdelt[0] > 0: m = m[...,:,::-1]
			if flip[0]: m = m[...,::-1,:]
			if flip[1]: m = m[...,:,::-1]
			# Update stamp list
			mlist[i] = m
		wcslist = [m.wcs for m in mlist]
		m = enmap.samewcs(np.asarray(mlist),mlist[0])
		if args.stamps is None:
			m, wcslist = m[0], None
		# Stack
		if args.tile is not None:
			toks = parse_list(args.tile, int)
			nrow = toks[0] if len(toks) > 0 else -1
			ncol = toks[1] if len(toks) > 1 else -1
			dims = (parse_list(args.tile_dims, int) or None) if args.tile_dims else None
			if dims is None and args.rgb: dims = list(range(m.ndim-3))
			m = hwstack(hwexpand(m, nrow, ncol, args.tile_transpose, dims=dims))[None]
		# Flatten pre-dimensions
		mf = m.reshape((-1,)+m.shape[-2:])
		# Mask bad data
		if args.mask is not None:
			mf  = mf*1.0 # Make the array floating point
			if not np.isfinite(args.mask): bad = np.abs(mf)==args.mask
			else: bad = np.abs(mf-args.mask)<=args.mask_tol
			if args.rgb: bad[:] = np.all(bad,0)
			mf[bad] = np.nan
		# Done
		if not return_info: return mf
		else:
			info = bunch.Bunch(fname=name, ishape=m.shape, wcslist=wcslist)
			return mf, info

def extract_stamps(map, args):
	"""Given a map, extract a set of identically sized postage stamps based on
	args.stamps. Returns a new map consisting of a stack of these stamps, along
	with a list of each of these' wcs object."""
	if args.stamps is None: return [map]
	# Stamps specified by format srcfile[:size[:nmax]], where the srcfile has
	# lines of [ra, dec] in degrees
	toks = args.stamps.split(":")
	# Read in our positions, optionally truncating the list
	srcs = np.loadtxt(toks[0], usecols=[0,1], ndmin=2).T[1::-1]*utils.degree
	size = int(toks[1]) if len(toks) > 1 else 16
	nsrc = int(toks[2]) if len(toks) > 2 else len(srcs.T)
	srcs = srcs[:,:nsrc]
	# Convert to pixel coordinates of corners
	pix  = np.round(map.sky2pix(srcs)-0.5*size).astype(int)
	# Extract stamps
	return map.stamps(pix.T, size, aslist=True)

def get_cache(cache, key, fun):
	if cache is None: return fun()
	if key not in cache: cache[key] = fun()
	return cache[key]

def draw_map_field(map, args, crange=None, return_layers=False, return_info=False, printer=noprint, cache=None):
	"""Draw a single map field, resulting in a single image. Adds a coordinate grid
	and lables as specified by args. If return_layers is True, an array will be
	returned instead of an image, wich each entry being a component of the image,
	such as the base image, the coordinate grid, the labels, etc. If return_bounds
	is True, then the """
	map, color = prepare_map_field(map, args, crange, printer=printer)
	tag    = (tuple(map.shape), map.wcs.to_header_string(), repr(args))
	layers = []
	names  = []
	yoff   = map.shape[-2]
	# Image layer
	if not args.no_image:
		with printer.time("to image", 3):
			img = PIL.Image.fromarray(utils.moveaxis(color,0,2)).convert('RGBA')
		layers.append((img,[[0,0],img.size]))
		names.append("img")
	# Contours
	if args.contours:
		with printer.time("draw contours", 3):
			contour_levels = calc_contours(crange, args)
			cimg = draw_contours(map, contour_levels, args)
			layers.append((cimg, [[0,0],cimg.size]))
			names.append("cont")
	# Annotations
	if args.annotate:
		with printer.time("draw annotations", 3):
			def get_aimg():
				annots = parse_annotations(args.annotate)
				return draw_annotations(map, annots, args)
			aimg = get_cache(cache, ("annotate",tag), get_aimg)
			layers.append((aimg, [[0,0],aimg.size]))
			names.append("annot")
	# Coordinate grid
	if args.grid % 2:
		with printer.time("draw grid", 3):
			ginfo = get_cache(cache, ("ginfo",tag), lambda: calc_gridinfo(map.shape, map.wcs, args))
			grid  = get_cache(cache, ("grid", tag), lambda: draw_grid(ginfo, args))
			layers.append(grid)
			names.append("grid")
		if not args.nolabels:
			with printer.time("draw labels", 3):
				labels, bounds = get_cache(cache, ("labels",tag), lambda: draw_grid_labels(ginfo, args))
				yoff = bounds[1][1]
				layers.append((labels,bounds))
				names.append("tics")
	if args.colorbar % 2:
		with printer.time("draw colorbar", 3):
			bimg, bounds = draw_colorbar(crange, map.shape[-1], args)
			bounds[:,1] += yoff
			yoff = bounds[1,1]
			layers.append((bimg,bounds))
			names.append("colorbar")
	 # Possibly other stuff too, like point source circles
	 # or contours
	with printer.time("stack layers", 3):
		layers, bounds = standardize_images(layers)
		if not return_layers: layers = merge_images(layers)
	info = bunch.Bunch(bounds=bounds, names=names)
	if return_info: return layers, info
	else: return layers

def draw_colorbar(crange, width, args):
	col  = tuple([int(args.font_color[i:i+2],16) for i in range(0,len(args.font_color),2)])
	font = cgrid.get_font(args.font_size)
	fmt  = "%g"
	labels, boxes = [], []
	for val in crange:
		# Val could be a one-element array. In NumPy 1.25 this is not
		# acceptable to string formatters.

		try:
			val = val[0]
		except (TypeError, IndexError):
			pass
		
		labels.append(fmt % val)
		boxes.append(font.getbbox(labels[-1])[-2:])
	boxes = np.array(boxes,int)
	lw, lh = np.max(boxes,0)
	img    = PIL.Image.new("RGBA", (width, lh))
	draw   = PIL.ImageDraw.Draw(img)
	# Draw the labels on the image
	draw.text((lw-boxes[0,0], 0), labels[0], col, font=font)
	draw.text((width-lw, 0),      labels[1], col, font=font)
	# Draw the color bar itself
	bar_data    = np.zeros((lh,width-2*lw))
	bar_data[:] = np.linspace(0,1,bar_data.shape[-1])
	bar_col     = map_to_color(bar_data[None], [0,1], args)
	bar_img     = PIL.Image.fromarray(utils.moveaxis(bar_col,0,2)).convert('RGBA')
	# Overlay it on the output image
	img.paste(bar_img, (lw,0))
	bounds = np.array([[0,0],[width,lh]])
	return img, bounds

def draw_map_field_mpl(map, args, crange=None, printer=noprint):
	"""Render a map field using matplotlib. Less tested and
	maintained than draw_map_field, and supports fewer features.
	Returns an object one can call savefig on to draw."""
	map, color = prepare_map_field(map, args, crange, printer=printer)
	# Set up matplotlib. We do it locally here to
	# avoid having it as a dependency in general
	with printer.time("matplotplib", 3):
		import matplotlib
		matplotlib.use("Agg")
		from matplotlib import pyplot, ticker
		matplotlib.rcParams.update({'font.size': 10})
		dpi, pad = args.mpl_dpi, args.mpl_pad
		winch, hinch = map.shape[2]/dpi, map.shape[1]/dpi
		fig  = pyplot.figure(figsize=(winch+pad,hinch+pad))
		box  = map.box()*180/np.pi
		pyplot.imshow(utils.moveaxis(color,0,2), extent=[box[0,1],box[1,1],box[1,0],box[0,0]])
		# Make conformal in center of image
		pyplot.axes().set_aspect(1/np.cos(np.mean(map.box()[:,0])))
		if args.grid % 2:
			ax = pyplot.axes()
			ticks = np.full(2,1.0)
			ticks[:] = parse_list(args.ticks)
			ax.xaxis.set_major_locator(ticker.MultipleLocator(ticks[1]))
			ax.yaxis.set_major_locator(ticker.MultipleLocator(ticks[0]))
			if args.subticks:
				ax.xaxis.set_minor_locator(ticker.MultipleLocator(args.sub))
				ax.yaxis.set_minor_locator(ticker.MultipleLocator(args.sub))
				pyplot.minorticks_on()
				pyplot.grid(True, which="major", linewidth=2)
				pyplot.grid(True, which="minor", linewidth=1)
			else:
				pyplot.grid(True)
		pyplot.tight_layout(pad=0.0,h_pad=0.0,w_pad=0.0)
	return pyplot
	#pyplot.savefig(oname,bbox_inches="tight",dpi=dpi)

def parse_range(desc,n):
	res = parse_list(desc, sep=":")[:n]
	return np.concatenate([res,np.repeat([res[-1]],n-len(res))])

def parse_list(desc, dtype=float, sep=","):
	if isinstance(desc, basestring):
		# Support strings of type "1.2,3.4,5,6"
		return [dtype(w) for w in desc.split(sep)]
	elif isinstance(desc, list) or isinstance(desc, tuple):
		# Support lists of numbers
		return [dtype(w) for w in desc]
	else:
		# Perhaps it's just a single plain number? If so,
		# convert it to the right dtype and return it as a 1-element list
		return [dtype(desc)]

def get_color_range(map, args):
	"""Compute an appropriate color bare range from map[:,ny,nx]
	given the args. Relevant members are range, min, max, quantile."""
	# Construct color ranges
	ncomp  = map.shape[0]
	crange = np.zeros((2,ncomp))+np.nan
	# Try explicit limits if given
	if args.range is not None:
		crange[1] = parse_range(args.range,ncomp)
		crange[0] = -crange[1]
	if args.min is not None: crange[0] = parse_range(args.min,ncomp)
	if args.max is not None: crange[1] = parse_range(args.max,ncomp)
	# Fall back on quantile otherwise
	if np.any(np.isnan(crange)):
		vals = np.sort(map[np.isfinite(map)])
		n    = len(vals)
		if n == 0: return np.repeat(np.array([-1,1])[:,None], ncomp, -1)
		i    = min(n-1,int(round(n*args.quantile)))
		v1, v2 = vals[i], vals[n-1-i]
		# Avoid division by zero later, in case min and max are the same
		if v2 == v1: (v1,v2) = (v1-1,v2+1)
		crange[0,np.isnan(crange[0])] = v1
		crange[1,np.isnan(crange[1])] = v2
	return crange

def get_num_digits(n): return int(np.log10(n))+1
def split_file_name(fname):
	"""Split a file name into directory, base name and extension,
	such that fname = dirname + "/" + basename + "." + ext."""
	dirname  = os.path.dirname(fname)
	if len(dirname) == 0: dirname = "."
	base_ext = os.path.basename(fname)
	# Find the extension. Using the last dot does not work for .fits.gz.
	# Using the first dot in basename does not work for foo2.5_bar.fits.
	# Will therefore handle .gz as a special case.
	if base_ext.endswith(".gz"):
		dot = base_ext[:-3].rfind(".")
	else:
		dot  = base_ext.rfind(".")
	if dot < 0: dot = len(base_ext)
	base = base_ext[:dot]
	ext  = base_ext[dot+1:]
	return dirname, base, ext

def map_to_color(map, crange, args):
	"""Compute an [{R,G,B},ny,nx] color map based on a map[1 or 3, ny,nx]
	map and a corresponding color range crange[{min,max}]. Relevant args
	fields: color, method, rgb. If rgb is not true, only the first element
	of the input map will be used. Otherwise 3 will be used."""
	map = ((map.T-crange[0])/(crange[1]-crange[0])).T # .T ensures broadcasting for rgb case
	if args.reverse_color: map = 1-map
	if args.rgb: m_color = colorize.colorize(map,    desc=args.color, driver=args.method, mode=args.rgb_mode)
	else:        m_color = colorize.colorize(map[0], desc=args.color, driver=args.method)
	m_color = enmap.samewcs(np.rollaxis(m_color,2), map)
	return m_color

def calc_gridinfo(shape, wcs, args):
	"""Compute the points making up the grid lines for the given map.
	Depends on args.ticks and args.nstep."""
	ticks = np.full(2,1.0)
	ticks[:] = parse_list(args.ticks)
	try:               unit = float(args.tick_unit)
	except TypeError:  unit = 1.0
	except ValueError: unit = args.tick_unit
	return cgrid.calc_gridinfo(shape, wcs, steps=ticks, nstep=args.nstep, zenith=args.zenith, unit=unit, positive_ra=args.pos_ra)

def draw_grid(ginfo, args):
	"""Return a grid based on gridinfo. args.grid_color controls the color
	the grid will be drawn with."""
	grid = cgrid.draw_grid(ginfo, color=args.grid_color, width=args.grid_width)
	bounds = np.array([[0,0],ginfo.shape[::-1]])
	return grid, bounds

def draw_grid_labels(ginfo, args):
	"""Return an image with a coordinate grid, along with abounds of this
	image relative to the coordinate shape stored in ginfo. Depends
	on the following args members: args.font, args.font_size, args.font_color"""
	linfo = []
	for gi in [ginfo.lat, ginfo.lon]:
		linfo += cgrid.calc_label_pos(gi, ginfo.shape[::-1])
	canvas = PIL.Image.new("RGBA", ginfo.shape[::-1])
	labels, bounds = cgrid.draw_labels(canvas, linfo, fname=args.font, fsize=args.font_size, color=args.font_color, return_bounds=True)
	return labels, bounds

def calc_contours(crange, args):
	"""Returns a list of values at which to place contours based on
	the valure range of the map crange[{from,to}] and the contour
	specification in args.

	Contour specifications:
		base:step or val,val,val...
	base: number
	step: number (explicit), -number (relative)
	"""
	if args.contours is None: return None
	def setup_uniform(vals):
		if len(vals) == 1:
			base, step = 0, vals[0]
		else:
			base, step = vals[:2]
		if step < 0:
			step = (crange[1]-crange[0])/(-step)
		# expand to fill crange
		a = int(np.ceil ((crange[0]-base)/step))
		b = int(np.floor((crange[1]-base)/step))+1
		return np.arange(a,b)*step + base
	# The input can either be a string or a list of numbers.
	# If it's a string, then it's interpreted based on whether
	# it contains commas or colons. If it's a list, then the
	# type is conrolled by args.contour_type
	if isinstance(args.contours, basestring):
		vals = args.contours.split(",")
		if len(vals) > 1:
			return np.array([float(v) for v in vals if len(v) > 0])
		else:
			return setup_uniform([float(tok) for tok in args.contours.split(":")])
	else:
		vals = parse_list(args.contours, float)
		if   args.contour_type == "list":    return np.array(vals)
		elif args.contour_type == "uniform":
			return setup_uniform(vals)
		else: raise ValueError("Unknown contour type '%s'" % args.contour_type)

def draw_contours(map, contours, args):
	img   = PIL.Image.new("RGBA", map.shape[-2:][::-1])
	inds  = np.argsort(contours)
	cmap  = array_ops.find_contours(map[0], contours[inds])
	cmap  = contour_widen(cmap, args.contour_width)
	cmap -= 1
	# Undo sorting if we sorted
	if not np.allclose(inds, np.arange(len(inds))):
		mask = cmap>=0
		cmap[mask] = inds[cmap[mask]]
	cmap  = cmap.astype(float)
	# Make non-contour areas transparent
	cmap[cmap<0] = np.nan
	# Rescale to 0:1
	if len(contours) > 1:
		cmap /= len(contours)-1
	color = colorize.colorize(cmap, desc=args.contour_color, driver=args.method)
	return PIL.Image.fromarray(color).convert('RGBA')

def parse_annotations(afile):
	try:
		with open(afile,"r") as f:
			return [shlex.split(line) for line in f]
	except (IOError, TypeError): return afile

def draw_annotations(map, annots, args):
	"""Draw a set of annotations on the map. These are specified
	as a list of ["type",param,param,...]. The recognized formats
	are:
		c[ircle] lat lon dy dx [rad [width [color]]]
		t[ext]   lat lon dy dx text [size [color]]
		l[ine]   lat lon dy dx lat lon dy dx [width [color]]
		r[ect]   lat lon dy dx lat lon dy dx [width [color]]
	dy and dx are pixel-unit offsets from the specified lat/lon.
	This is useful for e.g. placing text next to circles."""
	img  = PIL.Image.new("RGBA", map.shape[-2:][::-1])
	draw = PIL.ImageDraw.Draw(img, "RGBA")
	font = None
	font_size_prev = 0
	def topix(pos_off):
		unit = utils.degree if not wcsutils.is_plain(map.wcs) else 1.0
		pix = map.sky2pix(np.array([float(w) for w in pos_off[:2]])*unit)
		pix += np.array([float(w) for w in pos_off[2:]])
		return pix[::-1].astype(int)
	def skippable(x,y):
		rmax = args.annotate_maxrad
		if rmax == 0: return False
		return x <= -rmax or y <= -rmax or x >= map.shape[-1]-1+rmax or y >= map.shape[-2]-1+rmax
	for annot in annots:
		atype = annot[0].lower()
		color = "black"
		width = 2
		if atype in ["c","circle"]:
			x,y = topix(annot[1:5])
			if skippable(x,y): continue
			rad = 8
			if len(annot) > 5: rad   = int(annot[5])
			if len(annot) > 6: width = int(annot[6])
			if len(annot) > 7: color = annot[7]
			antialias = 1 if width < 1 else 4
			draw_ellipse(img,
					(x-rad,y-rad,x+rad,y+rad),
					outline=color,width=width, antialias=antialias)
		elif atype in ["l","line"] or atype in ["r","rect"]:
			x1,y1 = topix(annot[1:5])
			x2,y2 = topix(annot[5:9])
			nphi   = utils.nint(abs(360/map.wcs.wcs.cdelt[0]))
			x1, x2 = utils.unwind([x1,x2], nphi, ref=nphi//2)
			if skippable(x1,y1) and skippable(x2,y2): continue
			if len(annot) >  9: width = int(annot[9])
			if len(annot) > 10: color = annot[10]
			if atype[0] == "l":
				draw.line((x1,y1,x2,y2), fill=color, width=width)
			else:
				if x2 < x1: x1,x2 = x2,x1
				if y2 < y1: y1,y2 = y2,y1
				for i in range(width):
					draw.rectangle((x1+i,y1+i,x2-i,y2-i), outline=color)
		elif atype in ["t", "text"]:
			x,y  = topix(annot[1:5])
			if skippable(x,y): continue
			text = annot[5]
			size = 16
			if len(annot) > 6: size  = int(annot[6])
			if len(annot) > 7: color = annot[7]
			if font is None or size != font_size_prev:
				font = cgrid.get_font(size)
				font_size_prev = size
			tbox = font.getbbox(text)[-2:]
			draw.text((x-tbox[0]/2, y-tbox[1]/2), text, color, font=font)
		else:
			raise NotImplementedError
	return img

def standardize_images(tuples):
	"""Given a list of (img,bounds), composite them on top of each other
	(first at the bottom), and return the total image and its new bounds."""
	bounds_all = np.array([bounds for img, bounds in tuples])
	bounds_full= cgrid.calc_bounds(bounds_all, tuples[0][1][1])
	# Build canvas
	totsize = bounds_full[1]-bounds_full[0]
	res = []
	for img, bounds in tuples:
		# Expand to full size
		img_big = PIL.Image.new("RGBA", tuple(totsize))
		img_big.paste(img, tuple(bounds[0]-bounds_full[0]))
		res.append(img_big)
	return res, bounds_full

def merge_images(images):
	"""Stack all images into an alpha composite. Images must all have consistent
	extent before this. Use standardize_images to achieve this."""
	res = images[0]
	for img in images[1:]:
		res = PIL.Image.alpha_composite(res, img)
	return res

def merge_plots(plots):
	res = plots[0].copy()
	imgs, bounds = standardize_images([(plot.img, plot.info.bounds) for plot in plots])
	res.img         = imgs[0]
	res.info.bounds = bounds
	for img in imgs[1:]:
		res.img = PIL.Image.alpha_composite(res.img, img)
	return res

def prepare_map_field(map, args, crange=None, printer=noprint):
	if crange is None:
		with printer.time("ranges", 3):
			crange = get_color_range(map, args)
	if map.ndim == 2:
		map = map[None]
	if args.autocrop_each:
		map = enmap.autocrop(map, value=0)
	with printer.time("colorize", 3):
		color = map_to_color(map, crange, args)
	return map, color

def makefoot(n):
	b = np.full((2*n+1,2*n+1),1)
	b[n,n] = 0
	b = ndimage.distance_transform_edt(b)
	return b[1::2,1::2] < n

def contour_widen(cmap, width):
	if width <= 1: return cmap
	foot = makefoot(width)
	return ndimage.grey_dilation(cmap, footprint=foot)

def draw_ellipse(image, bounds, width=1, outline='white', antialias=1):
	"""Improved ellipse drawing function, based on PIL.ImageDraw.
	Improved from
	http://stackoverflow.com/questions/32504246/draw-ellipse-in-python-pil-with-line-thickness"""
	bounds = np.asarray(bounds)
	# Create small coordinate system around ellipse, with a
	# margin of width on each side
	esize  = bounds[2:]-bounds[:2] + 2*width
	ebounds= bounds - bounds[[0,1,0,1]] + width
	# Use a single channel image (mode='L') as mask.
	# The size of the mask can be increased relative to the imput image
	# to get smoother looking results. 
	mask = PIL.Image.new(size=tuple(esize*antialias), mode='L', color='black')
	draw = PIL.ImageDraw.Draw(mask)
	# draw outer shape in white (color) and inner shape in black (transparent)
	for offset, fill in (width/-2.0, 'white'), (width/2.0, 'black'):
		a = (ebounds[:2] + offset)*antialias
		b = (ebounds[2:] - offset)*antialias
		draw.ellipse([a[0],a[1],b[0],b[1]], fill=fill)
	# downsample the mask using PIL.Image.LANCZOS 
	# (a high-quality downsampling filter).
	mask = mask.resize(tuple(esize), PIL.Image.LANCZOS)
	# paste outline color to input image through the mask
	image.paste(outline, tuple(bounds[:2]-width), mask=mask)

def hwexpand(m, nrow=-1, ncol=-1, transpose=False, dims=None):
	"""Stack the maps in mflat[n,ny,nx] into a single flat map mflat[nrow,ncol,ny,nx]"""
	# Move tilable axes to just before the pixels
	if dims is None: dims = list(range(m.ndim-2))
	nflat = len(dims)
	npre  = m.ndim-nflat-2
	mflat = np.moveaxis(m, dims, [npre+i for i in range(nflat)])
	preshape = mflat.shape[:npre]
	# Flatten tilable axes
	mflat = mflat.reshape(preshape+(-1,)+mflat.shape[-2:])
	n, ny, nx = mflat.shape[-3:]
	if nrow < 0 and ncol < 0:
		ncol = int(np.ceil(n**0.5))
	if nrow < 0: nrow = (n+ncol-1)//ncol
	if ncol < 0: ncol = (n+nrow-1)//nrow
	if not transpose:
		omap = enmap.zeros(preshape+(nrow,ncol,ny,nx),mflat.wcs,mflat.dtype)
		omap.reshape(preshape+(-1,ny,nx))[...,:n,:,:] = mflat
	else:
		omap = enmap.zeros(preshape+(ncol,nrow,ny,nx),mflat.wcs,mflat.dtype)
		omap.reshape(preshape+(-1,ny,nx))[...,:n,:,:] = mflat
		omap = np.moveaxis(omap,(-4,-3),(-3,-4))
	return omap

def hwstack(mexp):
	nr,nc,ny,nx = mexp.shape[-4:]
	return np.moveaxis(mexp,-3,-2).reshape(mexp.shape[:-4]+(nr*ny,nc*nx))

class BackendError(BaseException): pass

def show(img, title=None, method="auto"):
	if   method == "tk":      show_tk(img, title=title)
	elif method == "qt":      show_qt(img, title=title)
	elif method == "wx":      show_wx(img, title=title)
	elif method == "ipython": show_ipython(img, title=title)
	elif method == "auto":
		# If matplotlib exists, use its perferences
		#try:
		#	import matplotlib
		#	backend = matplotlib.get_backend().lower()
		#	if    backend.startswith("tk"): return show_tk(img, title)
		#	elif  backend.startswith("qt"): return show_qt(img, title)
		#	elif  backend.startswith("wx"): return show_wx(img, title)
		#	elif "ipykernel" in backend: return show_ipython(img, title)
		#except ImportError: pass
		# If we get to this point, matplotlib couldn't tell us what to
		# do. Try them one by one in priority order
		try:
			# Only use ipython for graphical notebooks
			if get_ipython().__class__.__name__ in ["ZMQInteractiveShell", "Shell"]:
				return show_ipython(img, title=title)
		except (ImportError, NameError): pass
		try: return show_tk(img, title=title)
		except (ImportError, BackendError): pass
		try: return show_wx(img, title=title)
		except ImportError: pass
		try: return show_qt(img, title=title)
		except ImportError: pass
		# If we got here, nothing worked
		raise ImportError("Could not find any working display backends for enplot.show")

def show_ipython(img, title=None):
	from IPython.display import display
	for img_, title_ in _show_helper(img, title):
		display(img_)

def show_tk(img, title=None):
	from six.moves import tkinter
	from PIL import ImageTk
	class Displayer:
		def __init__(self):
			self.root    = tkinter.Tk()
			self.root.withdraw()
			self.windows = []
			self.nclosed = 0
		def add_window(self, img, title):
			window   = tkinter.Toplevel()
			window.minsize(img.width, img.height)
			window.title(title)
			canvas = tkinter.Canvas(window, width=img.width, height=img.height)
			canvas.pack()
			canvas.configure(background="white")
			photo  = ImageTk.PhotoImage(img)
			sprite = canvas.create_image(0, 0, image=photo, anchor=tkinter.NW)
			self.windows.append([window,canvas,photo,sprite])
			def closer():
				self.nclosed += 1
				window.destroy()
				if self.nclosed >= len(self.windows): self.root.destroy()
			window.protocol("WM_DELETE_WINDOW", closer)
	try:
		app = Displayer()
		for img_, title_ in _show_helper(img, title):
			app.add_window(img_, title_)
		app.root.mainloop()
	except tkinter.TclError: raise BackendError

def show_wx(img, title=None):
	import wx
	from PIL import Image
	class Panel(wx.Panel):
		def __init__(self, parent, id, img):
			wx.Panel.__init__(self, parent, id)
			self.SetBackgroundColour("white")
			# Make a non-transparend image with white background
			background = Image.new("RGB", img.size, (255,255,255))
			background.paste(img, mask=img.split()[3])
			wximg = wx.EmptyImage(*background.size)
			wximg.SetData(background.convert('RGB').tobytes())
			wxbmp = wx.BitmapFromImage(wximg)

			sizer = wx.BoxSizer()
			self.img_control = wx.StaticBitmap(self, -1, wxbmp, (0,0))
			sizer.Add(self.img_control, 1, wx.EXPAND)
			self.SetSizer(sizer)
			sizer.Fit(parent)
	app = wx.App(False)
	frames = []
	for img_, title_ in _show_helper(img, title):
		frame = wx.Frame(None, -1, title_, size=img_.size)
		Panel(frame,-1, img_)
		frame.Show(1)
		frames.append(frame)
	app.MainLoop()

def show_qt(img, title=None):
	from matplotlib.backends.backend_qt5 import QtCore, QtGui, QtWidgets
	from PIL.ImageQt import ImageQt
	import sys
	# Set up window
	class ImageWindow(QtWidgets.QMainWindow):
		def __init__(self, img, title):
			QtWidgets.QMainWindow.__init__(self)
			self.setWindowTitle(title)
			widget = QtWidgets.QWidget()
			self.setCentralWidget(widget)
			layout = QtWidgets.QVBoxLayout(widget)
			qimg   = QtGui.QImage(ImageQt(img))
			pixmap = QtGui.QPixmap(qimg)
			label = QtWidgets.QLabel()
			label.setPixmap(pixmap)
			label.adjustSize()
			layout.addWidget(label)
			self.resize(label.width(), label.height())
	app    = QtWidgets.QApplication([])
	windows= []
	for img_, title_ in _show_helper(img, title):
		window = ImageWindow(img_, title_)
		window.show()
		windows.append(window)
	app.exec_()

def _show_helper(img, title=None):
	res = []
	if isinstance(img, list):
		for i, im in enumerate(img):
			if isinstance(title, list): tit = title[i]
			else: tit = title
			res += _show_helper(im, tit)
		return res
	else:
		try:
			return [(img.img, (title or img.name))]
		except AttributeError:
			return [(img, (title or "plot"))]

def is_video_ext(ext):
	return ext in ["mp4", "mkv", "avi"]

class Writer:
	def __init__(self, **kwargs): pass
	def process(self, plot): raise NotImplementedError
	def close(self): pass
	def __enter__(self): return self
	def __exit__(self, type, value, traceback): self.close()

class PlotWriter(Writer):
	def __init__(self, **kwargs):
		self.vid_writer = VideoWriter(**kwargs)
	def process(self, plot, prefix=""):
		if plot.type == "vid":
			self.vid_writer.process(plot, prefix=prefix)
		elif plot.type == "pil":
			plot.img.save(prefix + plot.name)
		elif plot.type == "mpl":
			plot.img.savefig(prefix + plot.name,bbox_inches="tight",dpi=plot.dpi)
		else:
			raise ValueError("Unknown plot type '%s'" % plot.type)
	def close(self):
		self.vid_writer.close()

class VideoWriter(Writer):
	def __init__(self, codec="h264", crf=17, pix_fmt="yuv420p", fps=fractions.Fraction(30,1), **kwargs):
		self.fps    = fps
		self.crf    = crf
		self.pix_fmt= pix_fmt
		self.codec  = codec
		self.fname  = None
		self.output = None
	def process(self, plot, prefix=""):
		import av
		fname = prefix + plot.name
		if fname != self.fname:
			self.new(fname, plot.img)
		img    = merge_images([self.bg, plot.img])
		frame  = av.VideoFrame.from_image(img)
		packet = self.stream.encode(frame)
		self.output.mux(packet)
	def new(self, fname, img):
		import av
		# Close any existing output file
		self.close()
		# Set up a new one
		self.fname = fname
		self.output= av.open(self.fname, "w")
		stream         = self.output.add_stream("h264", self.fps)
		stream.width   = img.width
		stream.height  = img.height
		stream.pix_fmt = self.pix_fmt
		stream.options = {"crf": str(self.crf)}
		self.stream = stream
		# Background for transparent images
		self.bg = PIL.Image.new("RGBA", img.size, "WHITE")
	def close(self):
		if self.output is None: return
		packet = self.stream.encode(None)
		self.output.mux(packet)
		self.output.close()
		self.output = None



import cython
import numpy as np
cimport numpy as np
cimport cmisc
from libc.math cimport atan2
from libc.stdint cimport uintptr_t, int64_t

def alm2cl(ainfo, alm, alm2=None):
	"""Computes the cross power spectrum for the given alm and alm2, which
	must have the same dtype and broadcast. For example, to get the TEB,TEB
	cross spectra for a single map you would do
	 cl = ainfo.alm2cl(alm[:,None,:], alm[None,:,:])
	To get the same TEB,TEB spectra crossed with a different map it would
	be
	 cl = ainfo.alm2cl(alm1[:,None,:], alm2[None,:,:])
	In both these cases the output will be [{T,E,B},{T,E,B},nl]"""
	alm   = np.asarray(alm)
	alm2  = np.asarray(alm2) if alm2 is not None else alm
	# Unify dtypes
	dtype= np.result_type(alm, alm2)
	alm  = alm.astype (dtype, copy=False)
	alm2 = alm2.astype(dtype, copy=False)
	def getaddr(a): return a.__array_interface__["data"][0]
	# Broadcast alms. This looks scary, but just results in views of the original
	# arrays according to the documentation. Hence, it shouldn't use more memory.
	# The resulting arrays are non-contiguous, but each individual alm is still
	# contiguous. We set the writable flags not because we intend to write, but
	# to silience a false positive warning from numpy
	alm, alm2 = np.broadcast_arrays(alm, alm2)
	alm.flags["WRITEABLE"] = alm2.flags["WRITEABLE"] = True
	# I used to flatten here to make looping simple, but that caused a copy to be made
	# when combined with np.broadcast. So instead I will use manual raveling
	pshape = alm.shape[:-1]
	npre   = int(np.prod(pshape))
	cdef float[::1]  cl_single_sp, alm_single_sp1, alm_single_sp2
	cdef double[::1] cl_single_dp, alm_single_dp1, alm_single_dp2
	cdef int64_t[::1] mstart = np.ascontiguousarray(ainfo.mstart).view(np.int64)
	# A common use case is to compute TEBxTEB auto-cross spectra, where
	# e.g. TE === ET since alm1 is the same array as alm2. To avoid duplicate
	# calculations in this case we use a cache, which skips computing the
	# cross-spectrum of any given pair of arrays more than once.
	cache = {}
	if alm.dtype == np.complex64:
		cl = np.empty(alm.shape[:-1]+(ainfo.lmax+1,), np.float32)
		# We will loop over individual spectra
		for i in range(npre):
			I = np.unravel_index(i, pshape)
			# Avoid duplicate calculation
			key   = tuple(sorted([getaddr(alm[I]), getaddr(alm2[I])]))
			if key in cache:
				cl[I] = cache[key]
			else:
				alm_single_sp1 = np.ascontiguousarray(alm [I]).view(np.float32)
				alm_single_sp2 = np.ascontiguousarray(alm2[I]).view(np.float32)
				cl_single_sp = cl[I]
				cmisc.alm2cl_sp(ainfo.lmax, ainfo.mmax, &mstart[0], &alm_single_sp1[0], &alm_single_sp2[0], &cl_single_sp[0])
				cache[key] = cl[I]
		return cl
	elif alm.dtype == np.complex128:
		cl = np.empty(alm.shape[:-1]+(ainfo.lmax+1,), np.float64)
		# We will loop over individual spectra
		for i in range(npre):
			I = np.unravel_index(i, pshape)
			key   = tuple(sorted([getaddr(alm[I]), getaddr(alm2[I])]))
			if key in cache:
				cl[I] = cache[key]
			else:
				alm_single_dp1 = np.ascontiguousarray(alm [I]).view(np.float64)
				alm_single_dp2 = np.ascontiguousarray(alm2[I]).view(np.float64)
				cl_single_dp = cl[I]
				cmisc.alm2cl_dp(ainfo.lmax, ainfo.mmax, &mstart[0], &alm_single_dp1[0], &alm_single_dp2[0], &cl_single_dp[0])
				cache[key] = cl[I]
		return cl
	else:
		raise ValueError("Only complex64 and complex128 supported, but got %s" % str(alm.dtype))

# Repeated from utils to avoid dragging it in as a dependency
def nditer(shape):
	ndim = len(shape)
	I    = [0]*ndim
	while True:
		yield tuple(I)
		for dim in range(ndim-1,-1,-1):
			I[dim] += 1
			if I[dim] < shape[dim]: break
			I[dim] = 0
		else:
			break

def transpose_alm(ainfo, alm, out=None):
	"""In order to accomodate l-major ordering, which is not directoy
	supported by sharp, this function efficiently transposes Alm into
	Aml. If the out argument is specified, the transposed result will
	be written there. In order to perform an in-place transpose, call
	this function with the same array as "alm" and "out". If the out
	argument is not specified, then a new array will be constructed
	and returned."""
	if out is None: out = alm.copy()
	# Need this work array when the out and in arrays are the same.
	# Could avoid it for the non-inplace case, but for now always use it
	work = np.zeros(alm.shape[-1:], alm.dtype)
	cdef int64_t[::1] mstart = np.ascontiguousarray(ainfo.mstart).view(np.int64)
	cdef float[::1]  alm_sp, work_sp
	cdef double[::1] alm_dp, work_dp
	if out.dtype == np.complex128:
		work_dp = work.view(np.float64)
		for I in nditer(alm.shape[:-1]):
			alm_dp = alm[I].view(np.float64)
			cmisc.transpose_alm_dp(ainfo.lmax, ainfo.mmax, &mstart[0], &alm_dp[0], &work_dp[0])
			out[I] = work.view(np.complex128)
	elif out.dtype == np.complex64:
		work_sp = work.view(np.float32)
		for I in nditer(alm.shape[:-1]):
			alm_sp = alm[I].view(np.float32)
			cmisc.transpose_alm_sp(ainfo.lmax, ainfo.mmax, &mstart[0], &alm_sp[0], &work_sp[0])
			out[I] = work.view(np.complex64)
	else:
		raise ValueError("transpose_alm requires contiguous complex64 or complex128 arrays")
	return out

# I have C versions of this, but this one is more flexible and should be fast enough,
# given that it's just copying things around (and it's in cython anyway)
def transfer_alm(iainfo, ialm, oainfo, oalm=None, op=lambda a,b:b):
	"""Transfer alm from one layout to another."""
	if oalm is None:
		oalm = np.zeros(ialm.shape[:-1]+(oainfo.nelem,), ialm.dtype)
	lmax = min(iainfo.lmax, oainfo.lmax)
	mmax = min(iainfo.mmax, oainfo.mmax)
	if ialm.shape[:-1] != oalm.shape[:-1]:
		raise ValueError("ialm and oalm must agree on pre-dimensions")
	pshape = ialm.shape[:-1]
	npre = int(np.prod(pshape))
	# Numpy promotes uint64 to float64, so make an int64 view of mstart
	imstart = iainfo.mstart.view(np.int64)
	omstart = oainfo.mstart.view(np.int64)
	def transfer(dest, src, op): dest[:] = op(dest, src)
	for i in range(npre):
		I  = np.unravel_index(i, pshape)
		ia = ialm[I]; oa = oalm[I]
		for m in range(0, mmax+1):
			transfer(oa[omstart[m]+m*oainfo.stride:omstart[m]+(lmax+1)*oainfo.stride:oainfo.stride], ia[imstart[m]+m*iainfo.stride:imstart[m]+(lmax+1)*iainfo.stride:iainfo.stride], op)
	return oalm

def islastcontig(arr):
	return arr[(0,)*(arr.ndim-1)].flags["C_CONTIGUOUS"]
def aslastcontig(arr, dtype=None):
	arr = np.asarray(arr, dtype=dtype)
	if islastcontig(arr): return arr
	else: return np.ascontiguousarray(arr)

def lmul(ainfo, alm, lfun, out=None):
	import warnings
	# alm must be complex, lfun must be the corresponding real type
	alm  = np.asarray(alm)
	ctype= np.result_type(alm.dtype, 0j)
	rtype= np.zeros(1, alm.dtype).real.dtype
	# The arrays need to be contiguous along the last dimension
	alm  = aslastcontig(alm,  dtype=ctype)
	lfun = aslastcontig(lfun, dtype=rtype)
	if out is not None and not islastcontig(out) and out.dtype != ctype:
		raise ValueError("lmul's out argument must be contiguous along last axis, and have the same dtype as alm")
	# Are we doing a matmul?
	if lfun.ndim == 3 and alm.ndim == 2:
		if out is None: out = np.zeros(lfun.shape[:1]+alm.shape[1:], alm.dtype)
		if alm.dtype == np.complex128:  _lmatmul_dp(ainfo, alm, lfun, out)
		elif alm.dtype == np.complex64: _lmatmul_sp(ainfo, alm, lfun, out)
		else: raise ValueError("lmul requires complex64 or complex128 arrays")
	else:
		# Broadcast pre-dimensions, if they're compatible
		try:
			pre  = np.broadcast_shapes(alm.shape[:-1], lfun.shape[:-1])
		except ValueError:
			raise ValueError("lmul's alm and lfun's dimensions must either broadcast (when ignoring the last dimension), or have shape compatible with a matrix product (again ignoring the last dimension)")
		alm  = np.broadcast_to(alm,  pre+ alm.shape[-1:])
		lfun = np.broadcast_to(lfun, pre+lfun.shape[-1:])
		# Flatten, so the C code doesn't need to deal with variable dimensionality
		aflat= alm.reshape(-1,alm.shape[-1])
		lflat= lfun.reshape(-1,lfun.shape[-1])
		if out is None:
			out = aflat.copy()
		else:
			out = out.reshape(aflat.shape)
			out[:] = aflat
		if alm.dtype == np.complex128:  _lmul_dp(ainfo, out, lflat)
		elif alm.dtype == np.complex64: _lmul_sp(ainfo, out, lflat)
		else: raise ValueError("lmul requires complex64 or complex128 arrays")
		# Unflatten output
		out = out.reshape(pre + out.shape[-1:])
	return out

cdef _lmul_dp(ainfo, alm, lfun):
	cdef int64_t[::1] mstart = np.ascontiguousarray(ainfo.mstart).view(np.int64)
	cdef double[::1] _alm
	cdef const double[::1] _lfun
	for i in range(alm.shape[-2]):
		_alm  = alm [i].view(np.float64)
		_lfun = lfun[i].view(np.float64)
		cmisc.lmul_dp(ainfo.lmax, ainfo.mmax, &mstart[0], &_alm[0], lfun.shape[1]-1, &_lfun[0])

cdef _lmul_sp(ainfo, alm, lfun):
	cdef int64_t[::1] mstart = np.ascontiguousarray(ainfo.mstart).view(np.int64)
	cdef float  [::1] _alm
	cdef const float  [::1] _lfun
	for i in range(alm.shape[-2]):
		_alm  = alm [i].view(np.float32)
		_lfun = lfun[i].view(np.float32)
		cmisc.lmul_sp(ainfo.lmax, ainfo.mmax, &mstart[0], &_alm[0], lfun.shape[1]-1, &_lfun[0])

# Yuck! Cython is a pain. There must be an easier way to do this, but I didn't find it.
cdef _lmatmul_dp(ainfo, np.complex128_t[:,:] alm, np.float64_t[:,:,:] lfun, np.complex128_t[:,:] out):
	cdef int64_t[::1] mstart = np.ascontiguousarray(ainfo.mstart).view(np.int64)
	cdef np.uintp_t[::1] aptrs = np.empty(alm.shape[0], dtype=np.uintp)
	cdef np.uintp_t[::1] optrs = np.empty(out.shape[0], dtype=np.uintp)
	cdef np.uintp_t[::1] fptrs = np.empty(lfun.shape[0]*lfun.shape[1],dtype=np.uintp)
	cdef int i, j
	cdef int N = lfun.shape[0]
	cdef int M = lfun.shape[1]
	for i in range(N):
		for j in range(M):
			fptrs[i*M+j] = <np.uintp_t>&lfun[i,j,0]
	for i in range(N): optrs[i] = <np.uintp_t>&out[i,0]
	for j in range(M): aptrs[j] = <np.uintp_t>&alm[j,0]
	cmisc.lmatmul_dp(N, M, ainfo.lmax, ainfo.mmax, &mstart[0], <double**>&aptrs[0], lfun.shape[2]-1, <double**>&fptrs[0], <double**>&optrs[0])

cdef _lmatmul_sp(ainfo, np.complex64_t[:,:] alm, np.float32_t[:,:,:] lfun, np.complex64_t[:,:] out):
	cdef int64_t[::1] mstart = np.ascontiguousarray(ainfo.mstart).view(np.int64)
	cdef np.uintp_t[::1] aptrs = np.empty(alm.shape[0], dtype=np.uintp)
	cdef np.uintp_t[::1] optrs = np.empty(out.shape[0], dtype=np.uintp)
	cdef np.uintp_t[::1] fptrs = np.empty(lfun.shape[0]*lfun.shape[1],dtype=np.uintp)
	cdef int i, j
	cdef int N = lfun.shape[0]
	cdef int M = lfun.shape[1]
	for i in range(N):
		for j in range(M):
			fptrs[i*M+j] = <np.uintp_t>&lfun[i,j,0]
	for i in range(N): optrs[i] = <np.uintp_t>&out[i,0]
	for j in range(M): aptrs[j] = <np.uintp_t>&alm[j,0]
	cmisc.lmatmul_sp(N, M, ainfo.lmax, ainfo.mmax, &mstart[0], <float**>&aptrs[0], lfun.shape[2]-1, <float**>&fptrs[0], <float**>&optrs[0])



"""This is a convenience wrapper of pyfftw."""
from __future__ import print_function, division
import numpy as np, multiprocessing, os
from . import utils
engines = {}

# Define our engines. First a baseline numpy-based engine
class numpy_FFTW:
	"""Minimal wrapper of numpy in order to be able to provide it as an engine.
	Not a full-blown interface."""
	def __init__(self, a, b, axes=(-1), direction='FFTW_FORWARD', *args, **kwargs):
		self.a, self.b = a, b
		self.axes = axes
		self.direction = direction
	def __call__(self, normalise_idft=False):
		if self.direction == 'FFTW_FORWARD':
			if self.a.shape == self.b.shape:
				# Complex to complex
				self.b[:] = np.fft.fftn(self.a, axes=self.axes)
			else:
				# Real to complex
				self.b[:] = np.fft.rfftn(self.a, axes=self.axes)
		else:
			if self.a.shape == self.b.shape:
				# Complex to complex
				self.b[:] = np.fft.ifftn(self.a, axes=self.axes)
			else:
				self.b[:] = np.fft.irfftn(self.a, s=[self.b.shape[i] for i in self.axes], axes=self.axes)
			# Numpy already normalizes, so undo this if necessary
			if not normalise_idft:
				self.b *= np.prod([self.b.shape[i] for i in self.axes])

class ducc_FFTW:
	"""Minimal wrapper of ducc in order to be able to provide it as an engine.
	Not a full-blown interface."""
	def __init__(self, a, b, axes=(-1,), direction='FFTW_FORWARD', threads=1, *args, **kwargs):
		self.a, self.b = np.asarray(a), np.asarray(b)
		self.axes = tuple(axes)
		self.direction = direction
		self.threads   = threads
	def do_dct(self, kind, *args, **kwargs):
		# Expect format of type FFTW_REDFT00
		name = {"REDFT":"DCT","RODFT":"DST"}[kind[5:10]]
		num  = {"00":1, "10": 2, "01": 3, "11":4}[kind[10:12]]
		if   name == "DCT": return ducc0.fft.dct(*args, type=num, **kwargs)
		elif name == "DST": return ducc0.fft.dst(*args, type=num, **kwargs)
	def __call__(self, normalise_idft=False):
		if self.direction == 'FFTW_FORWARD':
			if self.a.shape == self.b.shape:
				# Complex to complex
				ducc0.fft.c2c(self.a, axes=self.axes, out=self.b, nthreads=self.threads)
			else:
				# Real to complex
				ducc0.fft.r2c(self.a, axes=self.axes, out=self.b, nthreads=self.threads)
		elif self.direction == "FFTW_BACKWARD":
			if self.a.shape == self.b.shape:
				# Complex to complex
				ducc0.fft.c2c(a=self.a, axes=self.axes, out=self.b, forward=False, inorm=2 if normalise_idft else 0, nthreads=self.threads)
			else:
				ducc0.fft.c2r(a=self.a, axes=self.axes, out=self.b, forward=False, lastsize=self.b.shape[self.axes[-1]], inorm=2 if normalise_idft else 0, nthreads=self.threads)
		elif _check_ducc_r2r(self.direction):
			# dct and dst are passed with a list with one entry per dimension of the transform,
			# but ducc doesn't support heterogeneous transforms like this, so just use the first element
			self.do_dct(self.direction[0], self.a, axes=self.axes, out=self.b, nthreads=self.threads)

def _check_ducc_r2r(direction):
	if isinstance(direction, str): return False
	for d in direction:
		if d != direction[0]:
			raise ValueError("ducc only supports homogeneous r2r transforms")
	return direction[0].startswith("FFTW_REDFT") or direction[0].startswith("FFTW_RODFT")

def numpy_empty_aligned(shape, dtype, n=None):
	"""This dummy function just skips the alignment, since numpy
	doesn't provide an easy way to get it."""
	return np.empty(shape, dtype)

class NumpyEngine: pass
numpy_engine = NumpyEngine()
numpy_engine.FFTW = numpy_FFTW
numpy_engine.empty_aligned = numpy_empty_aligned
engines["numpy"] = numpy_engine
engine = "numpy"

# Then optional, faster engines
try:
	import pyfftw
	engines["fftw"] = pyfftw
	engine = "fftw"
except ImportError: pass
else:
	# Is FFTW actually using intel MKL as a backend? Check if 1D FT crashes for 3D input.
	try:
		engines['fftw'].FFTW(np.zeros((1,1,1)), np.zeros((1,1,1), dtype=np.complex128),
			flags=['FFTW_ESTIMATE'], threads=1, axes=[-1])
	except RuntimeError as e:
		engines['intel'] = engines.pop('fftw')
		engine = 'intel'
try:
	import pyfftw_intel as intel
	engines["intel"] = intel
	engine = "intel"
except ImportError: pass
# ducc is slower than intel, but can be faster than pyfftw
try:
	import ducc0
	class DuccEngine: pass
	ducc_engine = DuccEngine()
	ducc_engine.FFTW = ducc_FFTW
	ducc_engine.empty_aligned = numpy_empty_aligned
	engines["ducc"] = ducc_engine
	if engine != "intel": engine = "ducc"
except ImportError: pass

if len(engines) == 0:
	# This should not happen due to the numpy fallback
	raise ImportError("Could not find any fftw implementations!")
try:
	nthread_fft = int(os.environ['OMP_NUM_THREADS'])
except (KeyError, ValueError):
	nthread_fft=multiprocessing.cpu_count()
nthread_ifft=nthread_fft
default_flags=['FFTW_ESTIMATE']
alignment = 32

def set_engine(eng):
	global engine
	engine = eng

def get_engine(eng):
	return engine if eng == "auto" else eng

def fft(tod, ft=None, nthread=0, axes=[-1], flags=None, _direction="FFTW_FORWARD", engine="auto"):
	"""Compute discrete fourier transform of tod, and store it in ft. What
	transform to do (real or complex, number of dimension etc.) is determined
	from the size and type of tod and ft. If ft is left out, a complex transform
	is assumed. The optional nthread argument specifies the number of theads to
	use in the fft. The default (0) uses the value specified by the
	OMP_NUM_THREAD environment varible if that is specified, or the total number
	of cores on the computer otherwise."""
	tod  = asfcarray(tod)
	axes = utils.astuple(-1 if axes is None else axes)
	if tod.size == 0: return
	nt = nthread or nthread_fft
	if flags is None: flags = default_flags
	if ft is None:
		otype = np.result_type(tod.dtype,0j)
		ft  = empty(tod.shape, otype)
		tod = tod.astype(otype, copy=False)
	engine = get_engine(engine)
	if engine == 'intel':
		ft[:] = fft_flat(tod, ft, axes=axes, nthread=nt, flags=flags, _direction=_direction)
	else:
		plan = engines[engine].FFTW(tod, ft, flags=flags, threads=nt, axes=axes, direction=_direction)
		plan()
	return ft

def ifft(ft, tod=None, nthread=0, normalize=False, axes=[-1],flags=None, engine="auto"):
	"""Compute inverse discrete fourier transform of ft, and store it in tod. What
	transform to do (real or complex, number of dimension etc.) is determined
	from the size and type of tod and ft. The optional nthread argument specifies
	the number of theads to use in the fft. The default (0) uses the value specified
	by the OMP_NUM_THREAD environment varible if that is specified, or the total
	number of cores on the computer otherwise. By default this is not normalized,
	meaning that fft followed by ifft will multiply the data by the length of the
	transform. By specifying the normalize argument, you can turn normalization
	on, though the normalization step will not use paralellization."""
	ft   = asfcarray(ft)
	axes = utils.astuple(-1 if axes is None else axes)
	if ft.size == 0: return
	nt = nthread or nthread_ifft
	if flags is None: flags = default_flags
	if tod is None:	tod = empty(ft.shape, ft.dtype)
	engine = get_engine(engine)
	if engine == 'intel':
		tod[:] = ifft_flat(ft, tod, axes=axes, nthread=nt, flags=flags)
	else:
		plan = engines[engine].FFTW(ft, tod, flags=flags, direction='FFTW_BACKWARD',
			threads=nt, axes=axes)
		plan(normalise_idft=False)
	# I get a small, cumulative loss in amplitude when using
	# pyfftw's normalize function.. So normalize manually instead	
	if normalize: tod /= np.prod([tod.shape[i] for i in axes])
	return tod

def rfft(tod, ft=None, nthread=0, axes=[-1], flags=None, engine="auto"):
	"""Equivalent to fft, except that if ft is not passed, it is allocated with
	appropriate shape and data type for a real-to-complex transform."""
	tod  = asfcarray(tod)
	axes = utils.astuple(-1 if axes is None else axes)
	if ft is None:
		oshape = list(tod.shape)
		oshape[axes[-1]] = oshape[axes[-1]]//2+1
		dtype = np.result_type(tod.dtype,0j)
		ft = empty(oshape, dtype)
	return fft(tod, ft, nthread, axes, flags=flags, engine=engine)

def irfft(ft, tod=None, n=None, nthread=0, normalize=False, axes=[-1], flags=None, engine="auto"):
	"""Equivalent to ifft, except that if tod is not passed, it is allocated with
	appropriate shape and data type for a complex-to-real transform. If n
	is specified, that is used as the length of the last transform axis
	of the output array. Otherwise, the length of this axis is computed
	assuming an even original array."""
	ft   = asfcarray(ft)
	axes = utils.astuple(-1 if axes is None else axes)
	if tod is None:
		oshape = list(ft.shape)
		oshape[axes[-1]] = n or (oshape[axes[-1]]-1)*2
		dtype = np.zeros([],ft.dtype).real.dtype
		tod = empty(oshape, dtype)
	return ifft(ft, tod, nthread, normalize, axes, flags=flags, engine=engine)

def dct(tod, dt=None, nthread=0, normalize=False, axes=[-1], flags=None, type="DCT-I", engine="auto"):
	"""Compute discrete cosine transform of tod, and store it in dt. By
	default it will do a DCT-I trasnform, but this can be controlled with the type argument.
	Even the much less common discrete sine transforms are avialble by passing e.g. type="DST-I".
	Valid values are DCT-I, DCT-II, DCT-III, DCT-IV, DST-I, DST-II, DST-III and DST-IV,
	or the raw FFTW names the correspond to (e.g. FFTW_REDFT00). If dt is not passed, it
	will be allocated with the same shape and data type as tod.

	The optional nthread argument specifies the number of theads to use in the fft. The
	default (0) uses the value specified by the OMP_NUM_THREAD environment varible if that
	is specified, or the total number of cores on the computer otherwise.

	Note that DCTs and DSTs were only added to pyfftw in version 13.0. The function will
	fail with an Invalid scheme error for older versions.
	"""
	tod  = asfcarray(tod)
	type = _dct_names[type]
	axes = utils.astuple(-1 if axes is None else axes)
	if dt is None:
		dt = empty(tod.shape, tod.dtype)
	return fft(tod, dt, nthread=nthread, axes=axes, flags=flags, _direction=[type]*len(axes), engine=engine)

def idct(dt, tod=None, nthread=0, normalize=False, axes=[-1], flags=None, type="DCT-I", engine="auto"):
	"""Compute the inverse discrete cosine transform of dt, and store it in tod. By
	default it will do the inverse of a DCT-I trasnform, but this can be controlled with the type argument.
	Even the much less common discrete sine transforms are avialble by passing e.g. type="DST-I".
	Valid values are DCT-I, DCT-II, DCT-III, DCT-IV, DST-I, DST-II, DST-III and DST-IV,
	or the raw FFTW names the correspond to (e.g. FFTW_REDFT00). If tod is not passed, it
	will be allocated with the same shape and data type as tod.

	By the default an unnormalized transform is performed. Pass normalize=True to get an
	actual inverse transform. This divides by a factor of 2*N+d for each axis the transform
	is performed along, where N is the length of the axis and d is -1 for DCT-1, +1 for
	DST-I and 0 for all the others. Usually it's faster to compute this factor once and
	combine it with other scalar factors in your math than to let this function do it,
	which is why it's turned off by default.

	Note that this function already takes care of figuring out which transform is the
	appropriate inverse. E.g. the inverse of b = dct(a, type="DCT-II") is
	idct(b, type="DCT-II", normalize=True), not idct(b, type="DCT-III", normalize=True)
	even though DCT-III is the inverse of DCT-II.

	The optional nthread argument specifies the number of theads to use in the fft. The
	default (0) uses the value specified by the OMP_NUM_THREAD environment varible if that
	is specified, or the total number of cores on the computer otherwise.

	Note that DCTs and DSTs were only added to pyfftw in version 13.0. The function will
	fail with an Invalid scheme error for older versions."""
	dt   = asfcarray(dt)
	type = _dct_inverses[_dct_names[type]]
	off  = _dct_sizes[type]
	axes = utils.astuple(-1 if axes is None else axes)
	if tod is None:
		tod = empty(dt.shape, dt.dtype)
	fft(dt, tod, nthread=nthread, axes=axes, flags=flags, _direction=[type]*len(axes), engine=engine)
	if normalize: tod /= np.prod([2*(tod.shape[i]+off) for i in axes])
	return tod

_dct_names = {
		"DCT-I":   "FFTW_REDFT00",  "FFTW_REDFT00":"FFTW_REDFT00",
		"DCT-II":  "FFTW_REDFT10",  "FFTW_REDFT10":"FFTW_REDFT10",
		"DCT-III": "FFTW_REDFT01",  "FFTW_REDFT01":"FFTW_REDFT01",
		"DCT-IV":  "FFTW_REDFT11",  "FFTW_REDFT11":"FFTW_REDFT11",
		"DST-I":   "FFTW_RODFT00",  "FFTW_RODFT00":"FFTW_RODFT00",
		"DST-II":  "FFTW_RODFT10",  "FFTW_RODFT10":"FFTW_RODFT10",
		"DST-III": "FFTW_RODFT01",  "FFTW_RODFT01":"FFTW_RODFT01",
		"DST-IV":  "FFTW_RODFT11",  "FFTW_RODFT11":"FFTW_RODFT11",
}
_dct_inverses = {
		"FFTW_REDFT00":"FFTW_REDFT00", "FFTW_REDFT10":"FFTW_REDFT01",
		"FFTW_REDFT01":"FFTW_REDFT10", "FFTW_REDFT11":"FFTW_REDFT11",
		"FFTW_RODFT00":"FFTW_RODFT00", "FFTW_RODFT10":"FFTW_RODFT01",
		"FFTW_RODFT01":"FFTW_RODFT10", "FFTW_RODFT11":"FFTW_RODFT11",
}
_dct_sizes = {
		"FFTW_REDFT00":-1, "FFTW_REDFT10":0, "FFTW_REDFT01":0, "FFTW_REDFT11":0,
		"FFTW_RODFT00":+1, "FFTW_RODFT10":0, "FFTW_RODFT01":0, "FFTW_RODFT11":0,
}

def redft00(a, b=None, nthread=0, normalize=False, flags=None, engine="auto"):
	"""Old brute-force work-around for missing dcts in pyfftw. Can be
	removed when newer versions of pyfftw become common. It's not very
	fast, sadly - about 5 times slower than an rfft. Transforms along the last axis."""
	a = asfcarray(a)
	if b is None: b = empty(a.shape, a.dtype)
	n = a.shape[-1]
	tshape = a.shape[:-1] + (2*(n-1),)
	itmp = empty(tshape, a.dtype)
	itmp[...,:n] = a[...,:n]
	itmp[...,n:] = a[...,-2:0:-1]
	otmp = rfft(itmp, axes=[-1], nthread=nthread, flags=flags, engine=engine)
	del itmp
	b[...] = otmp[...,:n].real
	if normalize: b /= 2*(n-1)
	return b

def chebt(a, b=None, nthread=0, flags=None, engine="auto"):
	"""The chebyshev transform of a, along its last dimension."""
	b = redft00(a, b, nthread, normalize=True, flags=flags, engine=engine)
	b[1:-1] *= 2
	return b

def ichebt(a, b=None, nthread=0, engine="auto"):
	"""The inverse chebyshev transform of a, along its last dimension."""
	a = asfcarray(a).copy()
	a[1:-1] *= 0.5
	return redft00(a, b, nthread, engine=engine)

def fft_len(n, direction="below", factors=None):
	if factors is None: factors = [2,3,5,7,11,13]
	return utils.nearest_product(n, factors, direction)

def asfcarray(a):
	a = np.asarray(a)
	return np.asarray(a, np.result_type(a,0.0))

def empty(shape, dtype):
	return engines[engine].empty_aligned(shape, dtype=dtype, n=alignment)

def fftfreq(n, d=1.0, dtype=np.float64): return np.fft.fftfreq(n, d=d).astype(dtype, copy=False)
def rfftfreq(n, d=1.0, dtype=np.float64): return np.arange(n//2+1, dtype=dtype)/(n*d)

def ind2freq (n, i, d=1.0): return np.where(i < n/2, i, -n+i)/(d*n)
def int2rfreq(n, i, d=1.0): return i/(n*d)
def freq2ind(n, f, d=1.0):
	j = f*(d*n)
	return np.where(j >= 0, j, n+j)
def rfreq2ind(n, f, d=1.0): return f*(n*d)

def shift(a, shift, axes=None, nofft=False, deriv=None, engine="auto"):
	"""Shift the array a by a (possibly fractional) number of samples "shift"
	to the right, along the specified axis, which defaults to the last one.
	shift can also be an array, in which case multiple axes are shifted together."""
	a      = np.asanyarray(a)
	ca     = a+0j
	shift  = np.atleast_1d(shift)
	if axes is None: axes = range(-len(shift),0)
	axes   = utils.astuple(axes)
	fa = fft(ca, axes=axes, engine=engine) if not nofft else ca
	for i, ax in enumerate(axes):
		ax   %= ca.ndim
		freqs = fftfreq(ca.shape[ax])
		phase = np.exp(-2j*np.pi*freqs*shift[i])
		if deriv == i:
			phase *= -2j*np.pi*freqs
		fa   *= phase[(None,)*ax + (slice(None),) + (None,)*(a.ndim-ax-1)]
	if not nofft: ifft(fa, ca, axes=axes, normalize=True, engine=engine)
	else:	      ca = fa
	return ca if np.iscomplexobj(a) else ca.real

def resample(a, n, axes=None, nthread=0, engine="auto"):
	"""Given an array a, resize the given axes (defaulting to the last ones) to
	length n (tuple or int) using Fourier resampling. For example, if a has shape
	(2,3,4), then resample(a, 10, -1) has shape (2,3,10), and resample(a, (20,10), (0,2))
	has shape (20,3,10)."""
	a    = np.asarray(a)
	n    = utils.astuple(n)
	if axes is None:
		axes = [-len(n)+i for i in range(len(n))]
	if len(n) != len(axes):
		raise ValueError("Resize size n = %s does not match axes = %s" % (str(n),str(axes)))
	fa   = fft(a, axes=axes, nthread=nthread, engine=engine)
	norm = 1/np.prod([a.shape[ax] for ax in axes])
	fa   = resample_fft(fa, n, axes=axes, norm=norm)
	out  = ifft(fa, axes=axes, normalize=False, nthread=nthread, engine=engine)
	if not np.iscomplexobj(a): out = out.real
	return out

def resample_fft(fa, n, out=None, axes=-1, norm=1, op=lambda a,b:b):
	"""Given array fa[{dims}] which is the fourier transform of some array a,
	transform it so that that it corresponds to the fourier transform of
	a version of a with a different number of samples by padding or truncating
	the fourier space. The argument n controls the new number of samples. By
	default this is for the last axis, but this can be changed using the axes
	argument. Multiple axes can be resampled at once by specifying a tuple for
	axes and n.

	The resulting array is multiplied by the argument norm. This can be used
	for normalization purposes. If norm is 1, then the multiplication is skipped.

	The argument out can be used to specify an already allocated output array.
	If it is None (the default), then an array will be allocated automatically.
	Normally the output array is overwritten, but this can be controlled using
	the op argument, which should be a function (out,fa)->out"""
	fa = np.asanyarray(fa)
	# Support n and axes being either tuples or a single number,
	# and broadcast n to match axes
	axes = utils.astuple(axes)
	n    = np.zeros(len(axes),int)+n
	# Determine the shape of the output array
	oshape = list(fa.shape)
	for i, ax in enumerate(axes):
		oshape[ax] = n[i]
	oshape = tuple(oshape)
	# Check or allocate output array
	if out is None:
		out = np.zeros(oshape, fa.dtype)
	else:
		if out.shape != oshape:
			raise ValueError("out argument has wrong shape in resample. Expected %s but got %s" % (str(oshape), str(out.shape)))
	# This function is used to avoid paying the cost of multiplying by norm when it's one
	def transfer(dest, source, norm, op):
		if norm != 1: source = source*norm
		dest[:] = op(dest, source)
	# Loop over start and end blocks for all dimensions
	for bi, I in enumerate(utils.nditer([2]*len(axes))):
		sel = [slice(None) for n in oshape]
		for ai, ax in enumerate(axes):
			c = min(fa.shape[ax], oshape[ax])
			if I[ai] == 0: sel[ax] = slice(0,c//2)
			else:          sel[ax] = slice(-(c-c//2),None)
		sel = tuple(sel)
		transfer(out[sel], fa[sel], norm, op)
	return out

def interpol_nufft(a, inds, out=None, axes=None, normalize=True,
		periodicity=None, epsilon=None, nthread=None, nofft=False, complex=False):
	"""Given some array a[{pre},{dims}] interpolate it at the given
	inds[len(dims),{post}], resulting in an output with shape [{pre},{post}].
	The signal is assumed to be periodic with the size of a unless this is overridden
	with the periodicity argument, which should have an integer for each axis being
	transformed. Normally the last ndim = len(inds) axes of a are interpolated.
	This can be overridden with the axes argument.

	By default the interpolation is properly normalized. This can be turned off
	with the normalization argument, in which case the output will be too high
	by a factor of np.prod([a.shape[ax] for ax in axes]). If all axes are used,
	this simplifies to a.size"""
	# This function could be implemented as simply u2nu(fft(a),inds). The problem
	# with this is that a full fourier-array needs to be allocated. I can save
	# some memory by instead doing the fft per field, at the cost of it being
	# a bit hacky
	op = None if nofft else lambda a, h: fft(a, nthread=h.nthread, axes=h.axall)
	return u2nu(a, inds, out=out, axes=axes, periodicity=periodicity,
		epsilon=epsilon, nthread=nthread, normalize=normalize, complex=complex, op=op)

def u2nu(fa, inds, out=None, axes=None, periodicity=None, epsilon=None, nthread=None,
			normalize=False, forward=False, complex=True, op=None):
	"""Given complex fourier coefficients fa[{pre},{dims}] corresponding to
	some real-space array a, evaluate the real-space signal at the given
	inds[len(dims),{post}], resulting in a output with shape [{pre},{post}].

	Arguments:
	* fa: Array of equi-spaced fourier coefficients. Complex with shape [{pre},{dims}]
	* inds: Array of positions at which to evaluate the inverse Fourier transform
	    of fa. Real with shape [len(dims),{post}]
	* out: Array to write result to. Real or complex with shape [{pre},{post}].
	    Optional. Allocated if missing.
	* axes: Tuple of axes to perform the transform along. len(axes)=len(dims).
	    Optional. Defaults to the last len(dims) axes.
	* periodicity: Periodicity assumed in the Fourier transform. Tuple with length
	   len(dims). Defaults to the shape of the axes being transformed.
	* epsilon: The target relative accuracy of the non-uniform FFT. Defaults
	   to 1e-5 for single precision and 1e-12 for double precision. See the
	   ducc0.nufft documentation for details.
	* normalize: If True, the output is divided by prod([fa.shape[ax] for ax in axes]),
	   that is, the total number of elements in the transform. This normalization is
	   equivalent to that of ifft. Defaults to False.
	* forward: Controls the sign of the exponent in the Fourier transform. By default
	   a backwards transform (fourier to real) is performed. By passing forward=True,
	   you can instead regard fa as a real-sapce array and out as a non-equispaced
	   Fourier array.
	* complex: Only relevant if out=None. Controls whether out is allocated as a
	   real or complex array. Defaults to complex.
	"""
	h = _nufft_helper(fa, out, inds, axes=axes, nuout=True, periodicity=periodicity,
		epsilon=epsilon, nthread=nthread, normalize=normalize, complex=complex)
	if op is None: op = lambda fa, h: fa
	for uI, nuI in zip(h.uiter, h.nuiter):
		grid = op(h.u[uI],h).astype(h.ctype, copy=False)
		res  = ducc0.nufft.u2nu(grid=grid, coord=h.iflat, forward=forward,
			epsilon=h.epsilon, nthreads=h.nthread, periodicity=h.periodicity,
			fft_order=True)
		if not np.iscomplexobj(h.nu):
			res = res.real
		h.nu[nuI] = res.reshape(h.inds.shape[1:])
	if h.normalize:
		h.nu /= h.norm
	return h.nu

# FIXME: Check normalization
def nu2u(a, inds, out=None, oshape=None, axes=None, periodicity=None, epsilon=None, nthread=None,
			normalize=False, forward=False):
	h = _nufft_helper(out, a, inds, axes=axes, nuout=False, periodicity=periodicity, ushape=oshape,
		epsilon=epsilon, nthread=nthread, normalize=normalize, complex=complex)
	work = np.zeros(h.tshape, h.ctype)
	for uI, nuI in zip(h.uiter, h.nuiter):
		res = ducc0.nufft.nu2u(points=h.nu[nuI], coord=h.iflat, out=work,
			forward=forward, epsilon=h.epsilon, nthreads=h.nthread,
			periodicity=h.periodicity, fft_order=True)
		if not np.iscomplexobj(h.u):
			res = res.real
		h.u[uI] = res
	if h.normalize:
		h.u /= h.norm
	return h.u

def iu2nu(a, inds, out=None, oshape=None, axes=None, periodicity=None, epsilon=None, nthread=None,
			normalize=False, forward=False):
	"""The inverse of nufft/u2nu. Given non-equispaced samples a[{pre},{post}] and
	their coordinates inds[len(dims),{post}], calculates the equispaced
	Fourier coefficients out[{pre},{dims}] of a.

	Arguments:
	* a: Array of of non-equispaced values. Real or complex with shape [{pre},{post}]
	* inds: Coordinates of samples in a. Real with shape [len(dims),{post}].
	* out: Equispaced Fourier coefficients of a. Complex with shape [{pre},{dims}].
	    Optional, but if missing, the shape of the out array to allocate must be
	    specified using the oshape argument
	* oshape: Tuple giving the shape to use when allocating out (if it's not passed in).
	See u2nu for the meaning of the other arguments.
	"""
	h = _nufft_helper(out, a, inds, axes=axes, nuout=False, ushape=oshape,
		periodicity=periodicity, epsilon=epsilon, nthread=nthread,
		normalize=normalize, complex=complex)
	work   = np.zeros(h.tshape, h.ctype)
	def wzip(u): return u.reshape(-1).view(h.rtype)
	def wunzip(x): return x.view(h.ctype).reshape(h.tshape)
	def P(u): return ducc0.nufft.u2nu(grid=u, coord=h.iflat, forward=forward,
		epsilon=h.epsilon, nthreads=h.nthread, periodicity=h.periodicity, fft_order=True)
	def PT(nu):
		return ducc0.nufft.nu2u(points=nu, coord=h.iflat,
		out=work, forward=not forward,
		epsilon=h.epsilon, nthreads=h.nthread, periodicity=h.periodicity, fft_order=True)
	for uI, nuI in zip(h.uiter, h.nuiter):
		# Invert u2nu by finding the least-squares solution to
		#  a = u2nu(out, inds). Written linearly this is a = P out
		# with solution out = (P'P)"P'a. The CG solver wants real numbers, though,
		# so we hack around that with view
		# Set up the equation system. Our degrees of freedom are flattened real u
		b = wzip(PT(h.nu[nuI].reshape(-1)))
		def A(x): return wzip(PT(P(wunzip(x))))
		solver = utils.CG(A, b)
		while solver.err > h.epsilon:
			solver.step()
		res = wunzip(solver.x)
		if not np.iscomplexobj(h.u):
			res = res.real
		h.u[uI] = res
	if h.normalize:
		h.u *= h.norm
	return h.u

# FIXME: Check normalization
def inu2u(fa, inds, out=None, axes=None, periodicity=None, epsilon=None, nthread=None,
			normalize=False, forward=False, complex=True):
	h = _nufft_helper(fa, out, inds, axes=axes, nuout=True,
		periodicity=periodicity, epsilon=epsilon, nthread=nthread,
		normalize=normalize, complex=complex)
	work = np.zeros(h.tshape, h.ctype)
	def wzip(nu): return nu.view(h.rtype)
	def wunzip(x): return x.view(h.ctype)
	def P(nu): return ducc0.nufft.nu2u(points=nu, coord=h.iflat, out=work, forward=forward,
		epsilon=h.epsilon, nthreads=h.nthread, periodicity=h.periodicity, fft_order=True)
	def PT(u): return ducc0.nufft.u2nu(grid=u, coord=h.iflat, forward=not forward,
		epsilon=h.epsilon, nthreads=h.nthread, periodicity=h.periodicity, fft_order=True)
	for uI, nuI in zip(h.uiter, h.nuiter):
		# Invert nu2u by finding the least-squares solution to
		#  fa = nu2u(out, inds). Written linearly this is fa = P out
		# with solution out = (P'P)"P'fa
		b = wzip(PT(h.u[uI]))
		def A(x): return wzip(PT(P(wunzip(x))))
		solver = utils.CG(A, b)
		while solver.err > h.epsilon:
			solver.step()
		res = wunzip(solver.x)
		if not np.iscomplexobj(h.nu):
			res = res.real
		h.nu[nuI] = res.reshape(h.inds.shape[1:])
	if h.normalize:
		h.nu *= h.norm
	return h.nu

# Alternative nufft interface more in line with fft and curvedsky.
# TODO: Add proper docstrings here. Can I avoid lots of repetition?

def nufft(a, inds, out=None, oshape=None, axes=None, periodicity=None, epsilon=None, nthread=None, normalize=False, flip=False):
	"""Alias of iu2nu(..., forward=flip). This involves inverting a system with conjugate gradients"""
	return iu2nu(a, inds, out=out, oshape=oshape, axes=axes, periodicity=periodicity, epsilon=epsilon, nthread=nthread, normalize=normalize, forward=flip)

def inufft(fa, inds, out=None, axes=None, periodicity=None, epsilon=None, nthread=None, normalize=False, flip=False, complex=True, op=None):
	"""Alias of u2nu(..., forward=flip)"""
	return u2nu(fa, inds, out=out, axes=axes, periodicity=periodicity, epsilon=epsilon, nthread=nthread, normalize=normalize, forward=flip, complex=complex, op=op)

def nufft_adjoint(a, inds, out=None, oshape=None, axes=None, periodicity=None, epsilon=None, nthread=None, normalize=False, flip=False):
	"""Alias of nu2u(..., forward=not flip)"""
	return nu2u(a, inds, out=out, oshape=oshape, axes=axes, periodicity=periodicity, epsilon=epsilon, nthread=nthread, normalize=normalize, forward=not flip)

def inufft_adjoint(fa, inds, out=None, axes=None, periodicity=None, epsilon=None, nthread=None, normalize=False, flip=False, complex=True):
	"""Alias of inu2u(..., forward=not flip). This involves inverting a system with conjugate gradients"""
	return inu2u(fa, inds, out=out, axes=axes, periodicity=periodicity, epsilon=epsilon, nthread=nthread, normalize=normalize, forward=not flip)

# Incremental nufft. This is useful for low-latency interpolation, or interpolation
# where you don't know how many points there will be

# Must figure out how much can be factorized out here. Sadly it wasn't
# practicaly to use _nufft_helper here. Will wait with factorization until
# I implement nu2u_plan
class u2nu_plan:
	def __init__(self, fa, axes, periodicity=None, epsilon=None, nthread=None,
			normalize=False, forward=False, complex=True, op=None):
		# Will set up one plan per fft-field
		fa           = np.asarray(fa)
		self.axes    = utils.astuple(axes)
		self.shape   = fa.shape
		self.gshape  = tuple([self.shape[ax] for ax in self.axes])
		self.paxes   = tuple(utils.complement_inds(self.axes, fa.ndim))
		self.pshape  = tuple([self.shape[ax] for ax in self.paxes])
		if op is None: op = lambda fa: fa
		if periodicity is None: periodicity = self.gshape
		else: periodicity = np.zeros(len(self.axes),int)+periodicity
		self.nthread = nthread or nthread_fft
		self.plans = []
		for I in utils.nditer(self.shape, axes=self.paxes):
			faI        = op(fa[I])
			self.ctype = faI.dtype
			# Target accuracy
			if epsilon is None:
				epsilon = 1e-5 if self.ctype == np.complex64 else 1e-12
			plan       = ducc0.nufft.experimental.incremental_u2nu(
				grid=faI, epsilon=epsilon, nthreads=self.nthread, forward=forward,
				periodicity=periodicity, fft_order=True)
			self.plans.append(plan)
		self.epsilon = epsilon
		self.forward = forward
		self.dtype   = utils.real_dtype(self.ctype)
		self.ndim    = len(self.axes)
		self.complex = complex
		self.norm    = np.prod([fa.shape[ax] for ax in axes])
		self.normalize = normalize
	def eval(self, inds, out=None):
		inds  = np.asarray(inds, dtype=self.dtype)
		iflat = inds.reshape(self.ndim,-1).T
		if out is None:
			out = np.zeros(self.pshape+inds.shape[1:], self.ctype if self.complex else self.dtype)
		oflat = out.reshape(len(self.plans),iflat.shape[0])
		for i, plan in enumerate(self.plans):
			vals = self.plans[i].get_points(coord=iflat)
			if not self.complex: vals = vals.real
			oflat[i] = vals
			del vals
		if self.normalize:
			out /= self.norm
		return out

########### Helper functions ##############


def fft_flat(tod, ft, nthread=1, axes=[-1], flags=None, _direction="FFTW_FORWARD"):
	"""Workaround for intel FFTW wrapper. Flattens appropriate dimensions of
	intput and output arrays to avoid crash that otherwise happens for arrays with
	ndim > N + 1, where N is the dimension of the transform. If 'axes' correspond
	to the last dimensions of the arrays, the workaround is essentially free. If
	`axes` correspond to other axes, copies are made when reshaping the arrays."""
	shape_ft = ft.shape
	naxes = np.atleast_1d(axes).size
	axes_new = list(range(-1, -1 - naxes, -1))
	ft = utils.partial_flatten(ft, axes=axes, pos=0)
	tod = utils.partial_flatten(tod, axes=axes, pos=0)
	plan = engines[engine].FFTW(tod, ft, flags=flags, threads=nthread, axes=axes_new, direction=_direction)
	plan()
	ft = utils.partial_expand(ft, shape_ft, axes=axes, pos=0)
	return ft

def ifft_flat(ft, tod, nthread=1, axes=[-1], flags=None):
	"""Same workaround as fft_flat but now for the inverse transform."""
	shape_tod = tod.shape
	naxes = np.atleast_1d(axes).size
	axes_new = list(range(-1, -1 - naxes, -1))
	tod = utils.partial_flatten(tod, axes=axes, pos=0)
	ft = utils.partial_flatten(ft, axes=axes, pos=0)
	plan = engines[engine].FFTW(ft, tod, flags=flags, direction='FFTW_BACKWARD',
		threads=nthread, axes=axes_new)
	plan(normalise_idft=False)
	tod = utils.partial_expand(tod, shape_tod, axes=axes, pos=0)
	return tod

def _nufft_helper(u, nu, inds, axes=None, periodicity=None, epsilon=None,
		nuout=False, nthread=None, complex=True, normalize=False, ushape=None):
	"""Do the type checking etc. needed to prepare for our nufft operations.
	This is a lot of code, but the overhead is around 300 µs, plus any time
	needed to allocate the output array. So there's a bit of overhead, but
	not anything we can't live with, and the raw ducc interface is available
	for when this overhead is too much."""
	from . import bunch
	# Prepare arguments for nufft operations. Must ensure that
	# * inds → iflat[ndim,npoint] f32 or f64
	# * ctype = c64 or c128 based u or nu, priority to which one is output which must be specified
	u     = np.asarray(u)  if u  is not None else None
	nu    = np.asarray(nu) if nu is not None else None
	inds  = np.asarray(inds)
	# Are we single or double precision? This set of statements sets up a priority
	# order for which array to get the dtype from
	dtypes = [nu if nuout else u, u if not nuout else nu, inds, np.float64()]
	rtypes = [utils.real_dtype(d.dtype) for d in dtypes if d is not None]
	rtype  = [d for d in rtypes if d in [np.float32, np.float64]][0]
	ctype  = utils.complex_dtype(rtype)
	if ctype not in [np.complex64, np.complex128]:
		raise ValueError("only single and double precision supported")
	# Convert inds to the right dtype only if it has an invalid dtype
	if inds.dtype not in [np.float32,np.float64]:
		inds = inds.astype(rtype, copy=False)
	ndim  = inds.shape[0]
	# By default the last ndim dimensions are transformed
	if axes is None: axes = tuple(range(-ndim,0))
	axes = utils.astuple(axes)
	if len(axes) != ndim: raise ValueError("Number of axes to transform does not match len(inds)!")
	# Set up output array. This depends on which direction we're going
	odtype = ctype if complex else rtype
	if nuout:
		npre   = u.ndim-ndim
		if npre < 0:
			raise ValueError("uniform array must has at least as many dimensions as indexed by the first axis of inds!")
		pshape = utils.without_inds(u.shape, axes)
		if nu is None:
			# Output array. Allocating it like this lets it inherit any subclass of
			# inds, which is useful when interpolating an enmap with another enmap
			nu = np.zeros_like(inds, shape=pshape+inds.shape[1:], dtype=odtype)
		if nu.shape != pshape+inds.shape[1:]:
			raise ValueError("nu must have shape pshape+inds.shape[1:]")
	else:
		if u is None:
			if ushape is None: raise ValueError("Either the output uniformly sampled array or its shape must be provided")
			u  = np.zeros(ushape, dtype=odtype)
		npre   = u.ndim-ndim
		pshape = utils.without_inds(u.shape, axes)
		# Hard to do any more sanity checks here
	tshape = utils.only_inds(u.shape, axes)
	npoint = np.prod(tshape)
	# Periodicity of the full space. Allows us to support arrays that represent
	# a subset of a bigger, periodic array
	if periodicity is None: periodicity = tshape
	else: periodicity = np.zeros(ndim,int)+periodicity
	nthread = nthread or nthread_fft
	# Target accuracy
	if epsilon is None:
		epsilon = 1e-5 if ctype == np.complex64 else 1e-12
	# ducc wants just a single pre-dimension for inds, so flatten it.
	iflat = inds.reshape(ndim,-1).T
	# Do the actual looping
	other_axes = tuple(utils.complement_inds(axes, u.ndim))
	axall = tuple(range(ndim))
	norm  = np.prod([u.shape[ax] for ax in axes])
	uiter  = utils.nditer(u.shape, axes=other_axes)
	nuiter = utils.nditer(nu.shape[:npre])
	return bunch.Bunch(u=u, nu=nu, inds=inds, iflat=iflat,
		epsilon=epsilon, nthread=nthread, normalize=normalize, norm=norm,
		periodicity=periodicity, pshape=pshape, tshape=tshape, npoint=npoint,
		other_axes=other_axes, axall=axall, complex=complex, rtype=rtype,
		ctype=ctype, npre=npre, uiter=uiter, nuiter=nuiter)
